package Urb::Boundary;
use strict;
use warnings;
use Math::Trig qw /:radial :pi/;

=head1 NAME

Urb::Boundary - A division within a quadrilateral space

=head1 SYNOPSIS

Models a division within an Urb::Quad object and child quads that share this boundary

=head1 DESCRIPTION

A boundary is a division of a quad or the edge of the root quad, it is always a
straight line.  It has various leafnode quads attached to it.  The boundary
object is a list of references to these leafnodes in no particular order.

=head1 USAGE

Create a new boundary object:

  $boundary = new Urb::Boundary;

=cut

sub new
{
    my $class = shift;
    $class = ref $class || $class;
    my $self = [];
    bless $self, $class;
    return $self;
}

=pod

Attach a quad to this boundary, indicate which edge of the quad (0,1,2 or 3) is attached:

  $boundary->Add_Edge ($quad, 2);

=cut

# FIXME test needed
sub Add_Edge
{
    my $self = shift;
    my ($quad, $id_edge) = @_;
    push @{$self}, {quad => $quad, id_edge => $id_edge};
}

=pod

Query the Id of the quad of which this boundary is a division:

  $boundary->Id;

=cut

sub Id
{
    my $self = shift;
    return undef unless defined $self->[0];
    return $self->[0]->{quad}->Boundary_Id ($self->[0]->{id_edge});
}

=pod

Query the total length of this boundary:

  $boundary->Length_Total;

=cut

sub Length_Total
{
    my $self = shift;
    my $quad_parent = $self->[0]->{quad}->By_Id ($self->Id);
    return _distance ($quad_parent->Coordinate_a, $quad_parent->Coordinate_b);
}

=pod

Check some internal consistency:

  $boolean = $boundary->Validate_Ids;

=cut

sub Validate_Ids
{
    my $self = shift;
    my $id = $self->Id;
    for my $item (@{$self})
    {
        return 0 unless $id eq $item->{quad}->Boundary_Id ($item->{id_edge});
    }
    return 1;
}

=pod

Given two quads, find out how much they overlap on this boundary, if at all:

  $distance = $boundary->Overlap ($quad_a, $quad_b);

=cut

sub Overlap
{
    my $self = shift;
    my ($quad_a, $quad_b) = @_;
    my ($edge_a, $edge_b);
    for my $item (@{$self})
    {
        $edge_a = $item->{id_edge} if $item->{quad} == $quad_a;
        $edge_b = $item->{id_edge} if $item->{quad} == $quad_b;
    }
    return 0.0 unless (defined $edge_a and defined $edge_b);

    my $length_a = $quad_a->Length ($edge_a);
    my $length_b = $quad_b->Length ($edge_b);

    my $length = 0;
    my $d;
    $d = _distance ($quad_a->Coordinate ($edge_a), $quad_b->Coordinate ($edge_b));
    $length = $d if $d > $length;
    $d = _distance ($quad_a->Coordinate ($edge_a), $quad_b->Coordinate ($edge_b +1));
    $length = $d if $d > $length;
    $d = _distance ($quad_a->Coordinate ($edge_a +1), $quad_b->Coordinate ($edge_b));
    $length = $d if $d > $length;
    $d = _distance ($quad_a->Coordinate ($edge_a +1), $quad_b->Coordinate ($edge_b +1));
    $length = $d if $d > $length;

    return $length_a if $length <= $length_b;
    return $length_b if $length <= $length_a;

    return $length_a + $length_b - $length;
}

=pod

Query coordinates of the segment shared by two overlapping quads:

  ($xy1, $xy2) = $boundary->Coordinates ($quad_a, $quad_b);

=cut

sub Coordinates
{
    my $self = shift;
    my ($quad_a, $quad_b) = @_;
    my ($edge_a, $edge_b);
    for my $item (@{$self})
    {
        $edge_a = $item->{id_edge} if $item->{quad} == $quad_a;
        $edge_b = $item->{id_edge} if $item->{quad} == $quad_b;
    }
    return 0.0 unless (defined $edge_a and defined $edge_b);

    my $length_a = $quad_a->Length ($edge_a);
    my $length_b = $quad_b->Length ($edge_b);

    my $length = 0;
    my @coordinates;
    my $d;
    $d = _distance ($quad_a->Coordinate ($edge_a), $quad_b->Coordinate ($edge_b));
    if ($d > $length)
    {
        $length = $d;
        @coordinates = ($quad_a->Coordinate ($edge_a +1), $quad_b->Coordinate ($edge_b +1));
    }
    $d = _distance ($quad_a->Coordinate ($edge_a), $quad_b->Coordinate ($edge_b +1));
    if ($d > $length)
    {
        $length = $d;
        @coordinates = ($quad_a->Coordinate ($edge_a +1), $quad_b->Coordinate ($edge_b));
    }
    $d = _distance ($quad_a->Coordinate ($edge_a +1), $quad_b->Coordinate ($edge_b));
    if ($d > $length)
    {
        $length = $d;
        @coordinates = ($quad_a->Coordinate ($edge_a), $quad_b->Coordinate ($edge_b +1));
    }
    $d = _distance ($quad_a->Coordinate ($edge_a +1), $quad_b->Coordinate ($edge_b +1));
    if ($d > $length)
    {
        $length = $d;
        @coordinates = ($quad_a->Coordinate ($edge_a), $quad_b->Coordinate ($edge_b));
    }

    # edge_a is contained entirely within edge_b
    @coordinates = ($quad_a->Coordinate ($edge_a), $quad_a->Coordinate ($edge_a +1)) if $length <= $length_b;
    # edge_b is contained entirely within edge_a
    @coordinates = ($quad_b->Coordinate ($edge_b), $quad_b->Coordinate ($edge_b +1)) if $length <= $length_a;
    # otherwise edge_a and edge_b partially overlap so we want the two inner points

    my $rad_quads = _vector_to_angle (_subtract_vector ($quad_b->Centroid, $quad_a->Centroid));
    my $rad_edge =  _vector_to_angle (_subtract_vector (@coordinates));
    my $rad = $rad_edge - $rad_quads;
    $rad += pi2 if $rad < 0;
    return @coordinates if $rad > pi;
    return reverse (@coordinates);
}

=pod

Query bearing between centres of two quads:

  $radians = $boundary->Bearing ($quad_a, $quad_b);

=cut

sub Bearing
{
    my $self = shift;
    my ($quad_a, $quad_b) = @_;
    my ($edge_a, $edge_b);
    for my $item (@{$self})
    {
        $edge_a = $item->{id_edge} if $item->{quad} == $quad_a;
        $edge_b = $item->{id_edge} if $item->{quad} == $quad_b;
    }
    return undef unless (defined $edge_a and defined $edge_b);

    my $vector = _subtract_vector ($quad_b->Centroid, $quad_a->Centroid);
    return _angle_to_bearing (_vector_to_angle ($vector));
}

=pod

Get a list of all pairs of quads that have a segment that is part of this boundary:

  $pairs = $boundary->Pairs;

=cut

# FIXME this should return a list not a list reference

sub Pairs
{
    my $self = shift;
    my $pairs = [];
    return $pairs if (defined $self->Id and $self->Id =~ /^[abcd]$/);
    for my $index_a (0 .. scalar @{$self} -1)
    {
        for my $index_b ($index_a +1 .. scalar @{$self} -1)
        {
            my $overlap = $self->Overlap ($self->[$index_a]->{quad}, $self->[$index_b]->{quad});
            push @{$pairs}, [$self->[$index_a]->{quad}, $self->[$index_b]->{quad}]
                if $overlap > 0;
        }
    }
    return $pairs;
}

=pod

Get a list of all pairs as with Pairs(), but sorted by length of shared segment:

  $link = $boundary->Pairs_By_Length;

=cut

sub Pairs_By_Length
{
    my $self = shift;
    my @list = sort {$self->Overlap ($a->[0], $a->[1]) <=> $self->Overlap ($b->[0], $b->[1])}
        @{$self->Pairs};
    return [@list];
}

# some internal functions
# FIXME should use the nice versions of these functions in Urb::Polygon

sub _distance
{
     my ($a, $b) = @_;
     return sqrt (($a->[0] - $b->[0])**2 + ($a->[1] - $b->[1])**2)
}

sub _subtract_vector
{
    my ($a, $b) = @_;
    return [$a->[0] - $b->[0], $a->[1] - $b->[1]];
}

sub _vector_to_angle
{
    my $vector = shift;
    return [cartesian_to_cylindrical (@{$vector}, 0)]->[1];
}

sub _angle_to_bearing
{
    my $angle = shift;
    my $deg = 90 - Math::Trig::rad2deg ($angle);
    $deg += 360 if $deg < 0;
    $deg -= 360 if $deg > 360;
    return $deg;
}

1;
