package Urb::Dom;

use strict;
use warnings;

use Urb::Quad;

use vars qw /@ISA/;
@ISA = qw /Urb::Quad/;

=head1 NAME

Urb::Dom - Unit of Domestic architectural space

=head1 SYNOPSIS

A Domestic building subclass of L<Urb::Quad>

=head1 DESCRIPTION

=head1 USAGE

Create a new Urb::Dom object:

  $dom = new Urb::Dom;
  $child = new Urb::Quad ($dom);

=head2 Methods that operate on leaf nodes

Get a score for the perpendicularness of the corners of this quad:

  $score = $dom->Perpendicular (0.3);

parameter is sigma variation in radians from pi/2
set to 0.01 for very right-angle corners
set to 10 if you don't care

=cut

sub Perpendicular
{
    my $self = shift;
    my $sigma = shift;
    my @radian = map {$self->Angle ($_)} (0,1,2,3);
    my @score = map {_gaussian ($_, 1.0, 1.570796, $sigma)} @radian;
    return $score[0] * $score[1] * $score[2] * $score[3];
}

=pod

Get a score for the plan proportion of this quad:

  score = $dom->Proportion (1.5, 0.5);

Parameters are preferred aspect and sigma variation, i.e. 1 is a square, 2 is a
double square etc...

=cut

sub Proportion
{
    my $self = shift;
    my ($aspect, $sigma) = @_;
    my $self_aspect = $self->Aspect;
    return 1.0 if $self_aspect < $aspect;
    return _gaussian ($self_aspect, 1.0, $aspect, $sigma);
}

=pod

Get a score for the plan area of this quad:

  score = $dom->Size (12.0, 3.0);

Parameters are preferred area and sigma variation.

=cut

sub Size
{
    my $self = shift;
    my ($area, $sigma) = @_;
    return _gaussian ($self->Area, 1.0, $area, $sigma);
}

=pod

Get a score for the width of this quad:

  $score = $dom->Width (2.5, 0.2);

Parameters are preferred minimum width and sigma variation.

=cut

sub Width
{
    my $self = shift;
    my ($width, $sigma) = @_;
    my $self_width = $self->Length_Narrowest;
    return 1.0 if $self_width > $width;
    return _gaussian ($self_width, 1.0, $width, $sigma);
}

=pod

Get a score for the Uncrinkliness (Area/External wall):

  $score = $dom->Uncrinkliness ($graph, 2.5, 1);

Parameters are a Graph object, preferred uncrinkliness and sigma variation.

(uncrinkliness = 1 / crinkliness)

=cut

sub Uncrinkliness
{
    my $self = shift;
    my $graph = shift;
    my ($distance, $sigma) = @_;
    return _gaussian ($self->Crinkle_Distance ($graph), 1.0, $distance, $sigma);
}

=pod

Query if the quad is 'Outdoors':

  $boolean = $dom->Is_Outside;

=cut

sub Is_Outside
{
    my $self = shift;
    return 1 if $self->Type =~ /^o/i;
    return 0;
}

=pod

Query if the quad is 'Covered', returns 0 if any leafs above are 'outdoors':

  $boolean = $dom->Is_Covered;

=cut

# TODO tests
sub Is_Covered
{
    my $self = shift;
    my @above = $self->Above_Leafs;
    return 0 unless scalar @above;
    for my $above (@above)
    {
        return 0 if $above->Is_Outside;
    }
    return 1;
}

=pod

Query if the quad is 'supported', returns 1 if any leafs below are 'indoors':

  $boolean = $dom->Is_Supported;

=cut

# TODO tests
sub Is_Supported
{
    my $self = shift;
    my @below = $self->Below_Leafs;
    return 0 unless scalar @below;
    for my $below (@below)
    {
        return 1 unless $below->Is_Outside;
    }
    return 0;
}

=pod

Calculate the 'plot ratio', i.e an empty plot is 0.0, a single storey building
that fills the plot is 1.0, a four storey building that occupies half of the
plot is 2.0.

This is the gross internal floor area divided by the plot area.

  my $ratio = $dom->Plot_Ratio;

=cut

sub Plot_Ratio
{
    my $self = shift;
    return $self->Area_Internal / $self->Root->Lowest->Area;
}

=pod

Calculate the gross internal floor area, i.e. add up the area of all rooms on
all floors.

  my $area = $dom->Area_Internal;

=cut

sub Area_Internal
{
    my $self = shift;
    my $base = $self->Root->Lowest;
    my $area = 0.0;
    for my $leaf (map {$_->Leafs} ($base, $base->Levels_Above))
    {  
        $area += $leaf->Area unless $leaf->Type =~ /^o/i;
    }
    return $area;
}

=head2 Misc functions

These operate on the table output of the Ratios() method:

  my $ratios = $dom->Ratios;

Score based on the proportion of outdoor space:

  my $score = ratio_o ($ratios, 0.2, 0.2);

Parameters are the preferred proportion and sigma variation.

=cut

sub ratio_o
{
    my ($ratios, $ratio, $sigma) = @_;
    my $proportion_o;

    my @out = grep /^o/i, keys %{$ratios};
    my $out = shift @out;
    if (defined $out) { $proportion_o = $ratios->{$out} }
    else { $proportion_o = 0 }

    return _gaussian ($proportion_o, 1.0, $ratio, $sigma);
}

=pod

Score based on the proportion of a named type of non-outdoor space:

  my $score = ratio_type ($ratios, 'k', 0.1, 0.05);

Paraeters are the type, preferred proportion and sigma variation.

=cut

sub ratio_type
{
    my ($ratios, $type, $ratio, $sigma) = @_;
    my $proportion_type;
    my $proportion_non_o;

    my @key = grep /^$type/i, keys %{$ratios};
    my $key = shift @key;
    if (defined $key) { $proportion_type = $ratios->{$key} }
    else { $proportion_type = 0 }

    my @out = grep /^o/i, keys %{$ratios};
    my $out = shift @out;
    if (defined $out) { $proportion_non_o = 1 - $ratios->{$out} }
    else { $proportion_non_o = 1 }

    $proportion_non_o = 1 unless $proportion_non_o;
    return _gaussian ($proportion_type / $proportion_non_o, 1.0, $ratio, $sigma);
}

=head2 Methods that query the network graph

Query if the graph forms a credible circulation network:

  $boolean = $dom->Is_Connected ($graph);

Returns true if all /^c/ nodes are connected to each other.
In the process $graph is stripped of all but /^c/ nodes.

=cut

sub Is_Connected
{
    my $self = shift;
    my $graph = shift;
    warn 'Is_Connected(): no graph supplied' unless defined $graph;
    for my $vertex ($graph->vertices)
    {
        $graph->delete_vertex ($vertex) unless $vertex->Type =~ /^c/i;
    }
    return $graph->is_connected;
}

=pod

Query if any of the nodes in the graph have a circulation leaf above, also
returns 1 if there is no level above:

  $boolean = $dom->Is_Connected_Above ($graph);

=cut

# TODO tests
sub Is_Connected_Above
{
    my $self = shift;
    my $graph = shift;
    warn 'Is_Connected_Above(): no graph supplied' unless defined $graph;
    return 1 unless $self->Above;
    for my $above (map {$_->Above ? $_->Above->Leafs : ()} $graph->vertices)
    {
        next unless $above->Type =~ /^c/i;
        return 1;
    }
    return 0;
}

=pod

Calculate length of outside wall, not including private plot boundaries:

  $length = $dom->Length_Outside ($graph);

Adds together length of edges connected to outside quads + edges on root quad boundary that are not private.

Note: Outside quads that are covered count as _inside_.

=cut

sub Length_Outside
{
    my $self = shift;
    my $graph = shift;
    my $length = 0;
    for my $neighbour ($graph->neighbours ($self))
    {
        next unless $neighbour->Is_Outside;
        next if $neighbour->Is_Covered;
        $length += $graph->get_edge_weight ($self, $neighbour);
    }
    for my $id (0 .. 3)
    {
        my $external = $self->Boundary_Id ($id);
        next unless ($external =~ /^[abcd]$/);
        next if (defined $self->Root->Perimeter->{$external} and $self->Root->Perimeter->{$external} =~ /^private$/i);
        $length += $self->Length ($id);
    }
    return $length;
}

=pod

Calculate the uncrinkliness value of the quad:

  $crinkle_distance = $dom->Crinkle_Distance ($graph);

Area of quad / exterior length

=cut

sub Crinkle_Distance
{
    my $self = shift;
    my $graph = shift;
    my $length_outside = $self->Length_Outside ($graph);
    return 0 unless $length_outside;
    return return $self->Area / $length_outside;
}

=pod

Get a list of neighbour types for this quad:

  @list = $self->Neighbour_Types ($graph);

=cut

sub Neighbour_Types
{
    my $self = shift;
    my $graph = shift;
    return sort (map {$_->Type} $graph->neighbours ($self));
}

=pod

Get a list of useful circulation or access neighbour types:

  @list = $dom->Access ($graph);

i.e. is this quad connected to a circulation space?

=cut

sub Access
{
    my $self = shift;
    my $graph = shift;
    my @neighbour_types = $self->Neighbour_Types ($graph);
    if ($self->Type =~ /^k/i)
    {
        return grep /^(l|c)/i, @neighbour_types;
    }
    elsif ($self->Is_Outside)
    {
        return @neighbour_types;
    }
    elsif ($self->Type =~ /^c/i)
    {
        return @neighbour_types;
    }
    else
    {
        return grep /^c/i, @neighbour_types;
    }
}

=pod

Get a list of outdoor neighbours:

  @list = $dom->Access_Outside ($graph);

Returns useful o connections.

=cut

sub Access_Outside
{
    my $self = shift;
    my $graph = shift;
    return grep /^o/i, $self->Neighbour_Types ($graph);
}

=pod

Get a list of 'living' neighbours:

  @list = $dom->Access_Living ($graph);

Returns useful l connections.

=cut

sub Access_Living
{
    my $self = shift;
    my $graph = shift;
    return grep /^l/i, $self->Neighbour_Types ($graph);
}

=pod

Get a list of plot boundary edges for this quad if any:

  @list = $dom->Access_External;

Returns connections outside root quad.

=cut

sub Access_External
{
    my $self = shift;
    my @list;
    push @list, sort (grep /^[abcd]$/, map {$self->Boundary_Id ($_)} (0 .. 3));
    return @list;
}

=pod

Is this quad a suitable entrance transition?:

  $dom->Public_Access_Outside ($graph);

Succeeds if quad is a leafnode, is outside, has a non-private boundary edge and has an inside neighbour

=cut

sub Public_Access_Outside
{
    my $self = shift;
    my $graph = shift;
    # needs to be a leafnode
    return 0 if $self->Divided;
    # needs to be an 'outside' quad
    return 0 unless grep /^o/i, $self->Type;
    # needs to be bounded by a street
    return 0 unless $self->Public_Access;
    # needs a public inside quad neighbour
    return 0 unless grep /^[lck]/i, $self->Neighbour_Types ($graph);
    return 1;
}

=pod

Is this quad on the street edge?:

  $dom->Public_Access;

Succeeds if quad is on ground floor, is a leafnode, and has a non-private
external boundary edge i.e. is bounded by a street

=cut

sub Public_Access
{
    my $self = shift;
    # needs to be a leafnode
    return 0 if $self->Level;
    return 0 if $self->Divided;
    # needs to have public boundary edge
    for my $external ($self->Access_External)
    {
        next if (defined $self->Root->Perimeter->{$external} and $self->Root->Perimeter->{$external} =~ /^private$/i);
        return 1;
    }
    return 0;
}

=pod

Get the length of street boundary for this quad:

  $length = $dom->Public_Length;

Returns length of street edge on this quad

Note: only ground floor has street edge

=cut

sub Public_Length
{
    my $self = shift;
    my $length = 0.0;
    return $length if $self->Level;
    for my $external ($self->Access_External)
    {
        next if (defined $self->Root->Perimeter->{$external} and $self->Root->Perimeter->{$external} =~ /^private$/i);
        for my $index (0 .. 3)
        {
            next unless $self->Boundary_Id ($index) eq $external;
            $length += $self->Length ($index);
        }
    }
    return $length;
}

=pod

DEPRECATED
Count walls with light, i.e. for 'light on two sides' pattern:

  @list = $dom->Access_Light ($graph);

Returns number of light boundaries

=cut

sub Access_Light
{
    my $self = shift;
    my $graph = shift;
    my @list;
    push @list, $self->Access_Outside ($graph);
    for my $external ($self->Access_External)
    {
        next if (defined $self->Root->Perimeter->{$external} and $self->Root->Perimeter->{$external} =~ /^private$/i);
        push @list, $external;
    }
    return @list;
}

=pod

Assemble a table of relative areas of different 'types':

  $ratios = $dom->Ratios;

=cut

sub Ratios
{
    my $self = shift;
    my $ratios;
    my $area_all = 0.0;
    for my $leaf (map {$_->Leafs} ($self, $self->Levels_Above))
    {
        $area_all += $leaf->Area;
        my $type = $leaf->Type;
        $ratios->{$type} = 0.0 unless defined $ratios->{$type};
        $ratios->{$type} += $leaf->Area;
    }
    for my $key (keys %{$ratios}) { $ratios->{$key} /= $area_all }
    return $ratios;
}

=pod

Assemble a table of the numbers of each 'type' of quad:

  $count = $dom->Count;

=cut

sub Count
{
    my $self = shift;
    my $count;
    for my $leaf (map {$_->Leafs} ($self, $self->Levels_Above))
    {
        my $type = lc ($leaf->Type);
        $type =~ s/^(.).*/$1/;
        push @{$count->{$type}}, $leaf->Id;
    }
    return $count;
}

=pod

Remove superflous divisions in outdoor areas, this recurses, so it can be
applied to all quads by applying to the root quad of the lowest level:

  $dom->Merge_Divided;

=cut

sub Merge_Divided
{
    my $self = shift;
    $self->Left->Merge_Divided if ($self->Divided);
    $self->Right->Merge_Divided if ($self->Divided);
    $self->Above->Merge_Divided if (!$self->Parent and $self->Levels_Above);
    return if (defined $self->Root->Above and defined $self->Root->Above->By_Id ($self->Id));
    if ($self->Divided and $self->Left->Is_Outside and $self->Right->Is_Outside)
    {
        $self->Undivide;
        $self->Type ('O');
    }
    if ($self->Divided and $self->Left->Type =~ /^c/i and $self->Right->Type =~ /^c/i)
    {
        $self->Undivide;
        $self->Type ('C');
    }
}

=pod

Divide outdoor and circulation quads to match division of levels below

  $dom->Split_Undivided_Up;

=cut

sub Split_Undivided_Up
{
    my $self = shift;
    if (!$self->Divided and defined $self->Root->Below and defined $self->Root->Below->By_Id ($self->Id) and $self->Below->Divided)
    {
        if ($self->Is_Outside or $self->Type =~ /^c/i)
        {
            $self->Divide (0.5, 0.5);
            $self->Left->Type ($self->Type);
            $self->Right->Type ($self->Type);
        }
    }
    $self->Left->Split_Undivided_Up if ($self->Divided);
    $self->Right->Split_Undivided_Up if ($self->Divided);
    $self->Above->Split_Undivided_Up if (!$self->Parent and $self->Levels_Above);
}

=pod

Divide outdoor and circulation quads to match division of levels above

  $dom->Split_Undivided_Down

=cut

sub Split_Undivided_Down
{
    my $self = shift;
    if (!$self->Divided and defined $self->Root->Above and defined $self->Root->Above->By_Id ($self->Id) and $self->Above->Divided)
    {
        if ($self->Is_Outside or $self->Type =~ /^c/i)
        {
            $self->Divide (0.5, 0.5);
            $self->Left->Type ($self->Type);
            $self->Right->Type ($self->Type);
        }
    }
    $self->Left->Split_Undivided_Down if ($self->Divided);
    $self->Right->Split_Undivided_Down if ($self->Divided);
    $self->Below->Split_Undivided_Down if (!$self->Parent and $self->Levels_Below);
}

sub _gaussian
{
    # $x = input value
    # $a = scale (usually 1.0)
    # $b = centre of peak
    # $c = sigma
    my ($x, $a, $b, $c) = @_;
    my $e = 2.718281828;
    return $a * ($e **(0- (($x-$b)**2/(2*$c*$c))));
}

1;
