package Urb::Dom::Fitness;

use strict;
use warnings;
use Urb::Dom;

use Exporter;
use vars qw /@ISA @EXPORT_OK/;
@ISA = qw /Exporter/;

@EXPORT_OK = qw /debug cost_leaf quality_perpendicular quality_proportion quality_size quality_width/;

our $DEBUG = $ENV{DEBUG};

=head1 NAME

Urb::Dom::Fitness - Fitness assessor for Urb::Dom objects

=head1 SYNOPSIS

A class that can be used to calculate fitness for domestic spaces.

=head1 DESCRIPTION

The eventual intention is to make this a Fitness assessor for L<Algorithm::Evolutionary>

=head1 USAGE

Create a new Urb::Dom::Fitness object:

  $fitness = new Urb::Dom::Fitness;

=cut

sub new
{
    my $class = shift;
    $class = ref $class || $class;
    my $self = {};
    bless $self, $class;
    $self->initialize();
    return $self;
}

# FIXME test needed
sub initialize
{
    my $self = shift;
    $self->{_counter} = 0; 
    $self->{_cache} = {};
}

=pod

Calculate fitness for an Urb::Dom object:

  $score = $fitness->_apply ($dom);

=cut

# FIXME test needed
sub _apply
{
    my $self = shift;
    my $dom = shift;
    $dom->Fail_Reset;
    $dom->Clean_Cache;

    # some initial values for whole multilevel
    my $has_public_access_outside = 0;
    my $has_public_private = 0;
    my $has_living_access = 1;
    my $public_length_all = 0.0;
    my $public_length_outside = 0.0;

    my $value = 0.0;
    my $cost = 10.0 * $dom->Area;
    debug ('initial cost:', $cost);

    # calculation recurses all levels
    for my $level ($dom, $dom->Levels_Above)
    {
    my $level_id = scalar $level->Levels_Below;

    my $graph = $level->Graph;

    for my $leaf ($level->Leafs)
        { $cost += cost_leaf ($leaf) }
    debug ('cost with leafs:', $cost);

    for my $edge ($graph->edges)
        { $cost += cost_edge ($graph, $edge) }
    debug ('cost with edges:', $cost);

    for my $leaf ($level->Leafs)
        { $cost += cost_edge_outside ($leaf) }
    debug ('cost with outside edges:', $cost);

    for my $leaf ($level->Leafs)
    {
        my $quality = 1;
        my $rate;
        my $id = $leaf->Id;
        my $case = $leaf->Type;
           if ($case =~ /^o/i) { $rate = 10.0 }
        elsif ($case =~ /^c/i) { $rate = 50.0 }
                          else { $rate = 300.0 }
        debug ('leaf level:', $leaf->Level, 'id:', $leaf->Id, 'type:', $leaf->Type, 'rate:', $rate, 'area:', $leaf->Area,
            'width:', $leaf->Length_Narrowest, 'proportion:', $leaf->Aspect);

        my $factor = quality_perpendicular ($leaf);
        $dom->Fail ("$level_id/$id perpendicular") if $factor < 0.3;
        $quality *= $factor;
        debug ('  quality perpendicular:', $factor);

        $factor = quality_proportion ($leaf);
        $dom->Fail ("$level_id/$id proportion") if $factor < 0.3;
        $quality *= $factor;
        debug ('  quality proportion:', $factor);

        $factor = quality_size ($leaf);
        $dom->Fail ("$level_id/$id size") if $factor < 0.3;
        $quality *= $factor;
        debug ('  quality size:', $factor);

        $factor = quality_width ($leaf);
        $dom->Fail ("$level_id/$id width") if $factor < 0.3;
        $quality *= $factor;
        debug ('  quality width:', $factor);

        $factor = quality_uncrinkliness ($leaf, $graph);
        $dom->Fail ("$level_id/$id crinkliness") if $factor < 0.3;
        $quality *= $factor;
        debug ('  quality crinkliness:', $factor);

        $has_public_access_outside = 1 if $leaf->Public_Access_Outside ($graph);
        $has_public_private = 1 if $leaf->Type =~ /^b/i and $leaf->Public_Access and $leaf->Level == 0;
        $has_living_access = 0 if $leaf->Type =~ /^k/i and not $leaf->Access_Living ($graph);

        my $public_length = $leaf->Public_Length;
        $public_length_all += $public_length;
        $public_length_outside += $public_length if $leaf->Type =~ /^o/i;

        # returns useful neighbour types

        my $count_access = scalar ($leaf->Access ($graph));
        if ($count_access > 0)
        {
            $factor = 1;
        }
        else
        {
            $factor = 0.01;
            $dom->Fail ("$level_id/$id access");
        }
        $quality *= $factor;
        debug ('  quality access:', $factor);

        $value += $quality * $rate * $leaf->Area;
    }

    # Returns true if all /^c/ nodes are connected to each other
    # In the process $graph is stripped of all but /^c/ nodes

    $dom->Fail ('not connected') unless ($level->Is_Connected ($graph));
    $dom->Fail ('not connected above') unless ($level->Is_Connected_Above ($graph));

    }

    # Ratios() does recurse through all levels
    my $ratios = $dom->Ratios;
    debug ('ratios:', map {$_ .':', $ratios->{$_}} keys %{$ratios});

    my $factor = Urb::Dom::ratio_o ($ratios,      0.20, 0.35);
    $value *= $factor;
    debug ('ratio_o:', $factor);

    $factor = Urb::Dom::ratio_type ($ratios, 'c', 0.00, 0.20);
    $value *= $factor;
    debug ('ratio_c:', $factor);

    $factor = Urb::Dom::ratio_type ($ratios, 't', 0.10, 0.06);
    $value *= $factor;
    debug ('ratio_t:', $factor);

    $factor = Urb::Dom::ratio_type ($ratios, 'k', 0.15, 0.10);
    $value *= $factor;
    debug ('ratio_k:', $factor);

    $factor = Urb::Dom::ratio_type ($ratios, 'l', 0.30, 0.25);
    $value *= $factor;
    debug ('ratio_l:', $factor);

    $factor = Urb::Dom::ratio_type ($ratios, 'b', 0.45, 0.15);
    $value *= $factor;
    debug ('ratio_b:', $factor);

    # Count() does recurse through all levels
    my $count = $dom->Count;

    $dom->Fail ('no b') if (! defined $count->{b} or scalar @{$count->{b}} < 1);
    $dom->Fail ('no l') if (! defined $count->{l} or scalar @{$count->{l}} < 1);
    $dom->Fail ('no k') if (! defined $count->{k} or scalar @{$count->{k}} < 1);
    $dom->Fail ('no t') if (! defined $count->{t} or scalar @{$count->{t}} < 1);
    $dom->Fail ('no o') if (! defined $count->{o} or scalar @{$count->{o}} < 1);
    $dom->Fail ('all o') if (defined $count->{o} and scalar keys %{$count} == 1);
    $dom->Fail ('too many k') if (defined $count->{k} and scalar @{$count->{k}} > 1);

    # only trigger these once
    $dom->Fail ('no outside public access') unless $has_public_access_outside;
    $dom->Fail ('bedroom on street') if $has_public_private;
    $dom->Fail ('kitchen not next to living') unless $has_living_access;

    $factor = Urb::Dom::_gaussian ($public_length_outside / $public_length_all, 1.0, 0.25, 0.2);
    $value *= $factor;
    debug ('ratio_public_outside:', $factor);

    $factor = Urb::Dom::_gaussian ($dom->Plot_Ratio, 1.0, 1.0, 0.2);
    $value *= $factor;
    debug ('plot ratio:', $factor);

    # final calculation after recursing through all levels
    $value *= 0.1**(scalar ($dom->Failures));
    debug ('quality fail:', $_) for ($dom->Failures);
    debug ('quality value:', $value);

    return $value / $cost;
}

=head2 Functions for calculating quality of leaf nodes

  $factor = quality_perpendicular ($leaf);

=cut

# FIXME test needed
sub quality_perpendicular
{
    my $leaf = shift;
    # parameter is sigma variation in radians from pi/2
    my $case = $leaf->Type;
    if ($case =~ /^o/i) { return $leaf->Perpendicular (10.0) }
                   else { return $leaf->Perpendicular (0.2) }
}

=pod

  $factor = quality_uncrinkliness ($leaf);

=cut

# FIXME test needed
sub quality_uncrinkliness
{
    my $leaf = shift;
    my $graph = shift;
    # parameter is distance in metres and sigma variation
    my $case = $leaf->Type;
       if ($case =~ /^o/i and !$leaf->Is_Covered) { return 1 }
    elsif ($case =~ /^c/i) { return $leaf->Uncrinkliness ($graph, 2.5, 0.8) }
                      else { return $leaf->Uncrinkliness ($graph, 2.5, 0.8) }
}

=pod

  $factor = quality_proportion ($leaf);

=cut

# FIXME test needed
sub quality_proportion
{
    my $leaf = shift;
    # parameters are preferred aspect and sigma variation
    my $case = $leaf->Type;
       if ($case =~ /^o/i) { return $leaf->Proportion (1.5, 50.0) }
    elsif ($case =~ /^c/i) { return $leaf->Proportion (1.5, 2.5) }
    elsif ($case =~ /^t/i) { return $leaf->Proportion (1.5, 2.0) }
                      else { return $leaf->Proportion (1.5, 0.5) }
}

=pod

  $factor = quality_size ($leaf);

=cut

# FIXME test needed
sub quality_size
{
    my $leaf = shift;
    # parameters are preferred area and sigma variation
    my $case = $leaf->Type;
       if ($case =~ /^o/i) { return 1 }
    elsif ($case =~ /^c/i) { return $leaf->Size (0.0, 14.0) }
    elsif ($case =~ /^k/i) { return $leaf->Size (16.0, 10.0) }
    elsif ($case =~ /^l/i) { return $leaf->Size (20.0, 10.0) }
    elsif ($case =~ /^b/i) { return $leaf->Size (12.0, 5.0) }
    elsif ($case =~ /^t/i) { return $leaf->Size (7.0, 3.0) }
                      else { return 1 }
}

=pod

  $factor = quality_width ($leaf);

=cut

# FIXME test needed
sub quality_width
{
    my $leaf = shift;
    # parameters are preferred minimum width and sigma variation
    my $case = $leaf->Type;
       if ($case =~ /^o/i and !$leaf->Is_Covered and !$leaf->Is_Supported) { return 1 }
    elsif ($case =~ /^o/i) { return $leaf->Width (2.5, 0.7) }
    elsif ($case =~ /^c/i) { return $leaf->Width (2.4, 1.5) }
    elsif ($case =~ /^k/i) { return $leaf->Width (3.0, 0.7) }
    elsif ($case =~ /^l/i) { return $leaf->Width (4.0, 1.0) }
    elsif ($case =~ /^b/i) { return $leaf->Width (4.0, 1.0) }
    elsif ($case =~ /^t/i) { return $leaf->Width (3.0, 0.7) }
                      else { return $leaf->Width (2.5, 0.1) }
}

=head2 Functions for calculating cost of leaf nodes

  $cost = cost_leaf ($leaf);

=cut

sub cost_leaf
{
    my $leaf = shift;
    my $rate;
    my $case = $leaf->Type;
    if ($case =~ /^o/i) { $rate = 10.0 }
                   else { $rate = 200.0 }
    if ($case =~ /^o/i and $leaf->Is_Covered) { $rate += 100 }
    if ($case =~ /^o/i and $leaf->Is_Supported) { $rate += 100 }
    return $rate * $leaf->Area;
}

=pod

  $cost = cost_edge ($leaf);

=cut

sub cost_edge
{
    my $graph = shift;
    my $edge = shift;
    my $rate;

    if ($edge->[0]->Type =~ /^o/i and $edge->[1]->Type =~ /^o/i)
    { $rate = 0.0 }

    elsif ($edge->[0]->Type =~ /^[^o]/i and $edge->[1]->Type =~ /^[^o]/i)
    { $rate = 200.0 }

    else
    { $rate = 300.0 }

    my $length = $graph->get_edge_weight (@{$edge});
    $edge->[0]->Root->Lowest->Fail (scalar ($edge->[0]->Root->Levels_Below) .' / '. $edge->[0]->Id .' '. $edge->[1]->Id .' edge too short') if ($length < 1.25);
    $edge->[0]->Root->Lowest->Fail (scalar ($edge->[0]->Root->Levels_Below) .' / '. $edge->[0]->Id .' '. $edge->[1]->Id .' edge too long') if ($length > 6.0 and $rate > 0.0);
    return $rate * $length;
}

=pod

  $cost = cost_edge_outside ($leaf);

=cut

sub cost_edge_outside
{
    my $leaf = shift;
    my $rate;
    my $case = $leaf->Type;
    if ($case =~ /^o/i) { $rate = 50.0 }
                   else { $rate = 400.0 }

    my $length = 0.0;
    for my $index (0 .. 3)
    {
        if ($leaf->Boundary_Id ($index) =~ /^[abcd]$/)
        {
            my $length_temp = $leaf->Length ($index);
            $length += $length_temp;
            $leaf->Root->Lowest->Fail (scalar ($leaf->Root->Levels_Below) .' / '. $leaf->Id .' outside edge too short') if ($length_temp < 1.25);
            next if ($leaf->Type =~ /^o/i);
            $leaf->Root->Lowest->Fail (scalar ($leaf->Root->Levels_Below) .' / '. $leaf->Id .' outside edge too long') if ($length_temp > 6.0);
        }
    }

    return $rate * $length;
}

sub debug
{
    return unless $DEBUG;
    print STDERR join ' ', @_, "\n";
}

1;
