package Urb::Dom::Mutate;

use strict;
use warnings;

our $DEBUG = $ENV{DEBUG};

=head1 NAME

Urb::Dom::Mutate - Mutator for Urb::Dom objects

=head1 SYNOPSIS

A class that can be used as a mutator for domestic spaces.

=head1 DESCRIPTION

The eventual intention is to make this an Op for L<Algorithm::Evolutionary>

=head1 USAGE

Create a new Urb::Dom::Mutate object:

  $mutator = new Urb::Dom::Mutate;

=cut

# FIXME need tests
sub new
{
    my $class = shift;
    $class = ref $class || $class;
    my $self = {};
    bless $self, $class;
    return $self;
}

=pod

Mutate an L<Urb::Dom> object:

  $mutator->apply ($dom);

This applies a random mutation from the list below.

=cut

sub apply
{
    my $self = shift;
    my $dom_original = shift;
    my $dom = $dom_original->Clone;
    $dom->Clean_Cache;

    my @all = map {$_->Children} ($dom, $dom->Levels_Above);
    my $index = int rand (scalar @all);
    my $victim = $all[$index];
    print "mutation victim: ". $victim->Level .'/'. $victim->Id ."\n" if $DEBUG;
    if (!$victim->Parent and !$victim->Above)
    {
        print "  rootnode\n" if $DEBUG;
        my $case = int rand (7);
           if ($case == 0) {rotate ($victim)}
        elsif ($case == 1) {slide ($victim)}
        elsif ($case == 2) {swap ($victim)}
        elsif ($case == 3) {undivide ($victim)}
        elsif ($case == 4) {level_add ($victim)}
        elsif ($case == 5) {level_delete ($victim)}
        elsif ($case == 6) {level_swap ($victim)}
    }
    elsif ($victim->Divided)
    {
        print "  branchnode\n" if $DEBUG;
        my $case = int rand (4);
           if ($case == 0) {rotate ($victim)}
        elsif ($case == 1) {slide ($victim)}
        elsif ($case == 2) {swap ($victim)}
        elsif ($case == 3) {undivide ($victim)}
    }
    else
    {
        print "  leafnode\n" if $DEBUG;
        my $case = int rand (2);
           if ($case == 0) {type ($victim)}
        elsif ($case == 1) {divide ($victim)}
    }
    # TODO mutations to swap, add and remove levels
    for my $branch (reverse $dom->Branches)
        { $branch->Collapse (0.9)}
    return $dom;
}

sub apply_rearrange
{
    my $self = shift;
    my $dom_original = shift;
    my $dom = $dom_original->Clone;
    $dom->Clean_Cache;

    my @all = map {$_->Branches} ($dom, $dom->Levels_Above);
    my $index = int rand (scalar @all);
    my $victim = $all[$index];
    print "mutation victim: ". $victim->Level .'/'. $victim->Id ."\n" if $DEBUG;
    my $case = int rand (3);
       if ($case == 0) {rotate ($victim)}
    elsif ($case == 1) {slide ($victim)}
    elsif ($case == 2) {swap ($victim)}
    return $dom;
}

=head2 mutation functions for branchnodes

Randomise the 'rotation' attribute:

  rotate ($dom);

=cut

sub rotate
{
    my $branchnode = shift;
    my $case = int rand (3);
       if ($case == 0) {$branchnode->Rotate}
    elsif ($case == 1) {$branchnode->Unrotate}
    elsif ($case == 2) {$branchnode->Rotate; $branchnode->Rotate}
    $branchnode->Straighten_Recursive (int rand (4));
    print "    rotate\n" if $DEBUG;
}

=pod

Randomise the position of a division of a branchnode dom object:

  slide ($dom);

=cut

sub slide
{
    my $branchnode = shift;
    # try and place new division location not near the ends
    my $length = $branchnode->Length ($branchnode->Rotation || 0);
    return unless $length;
    my $offset = 1.25 / $length;
    my $division = $offset + ((1 - $offset - $offset) * rand);

    $branchnode->Divide ($division, $division);
    $branchnode->Straighten_Recursive (int rand (4));
    print "    slide $division\n" if $DEBUG;
}

=pod

Exchange the left and right children of a branchnode

  swap ($dom);

=cut

sub swap
{
    my $branchnode = shift;
    $branchnode->Swap;
    $branchnode->Straighten_Recursive (int rand (4));
    print "    swap\n" if $DEBUG;
}

=pod

Remove all children of a branchnode and assign a random type:

  undivide ($dom);

=cut

sub undivide
{
    my $branchnode = shift;
    $branchnode->Undivide;
    type ($branchnode);
    print "    undivide\n" if $DEBUG;
}

=pod

Add a duplicate level above this, noop unless top level.

  level_add ($dom);

Delete all levels above this:

  level_delete ($dom);

Swap this level with the level above:

  level_swap ($dom);

=cut

sub level_add
{
    my $rootnode = shift;
    $rootnode->Clone_Above;
    print "    level_add\n" if $DEBUG;
}

sub level_delete
{
    my $rootnode = shift;
    return unless defined $rootnode->Below;
    $rootnode->Below->Del_Above;
    print "    level_delete\n" if $DEBUG;
}

sub level_swap
{
    my $rootnode = shift;
    return unless defined $rootnode->Above;
    $rootnode->Swap_Above;
    print "    level_swap\n" if $DEBUG;
}

=head2 mutation functions for leafnodes

Assign a random type to a leafnode:

  type ($dom);

=cut

sub type
{
    my $leafnode = shift;
    my @types = qw/T C K L B O/;
    $leafnode->Type ($types[int rand (scalar @types)]);
    print "    type\n" if $DEBUG;
}

=pod

Split a leafnode and turn into a branchnode, assign current type to one child
and a random type to the other:

  divide ($dom);

=cut

sub divide
{
    my $leafnode = shift;
    slide ($leafnode);
    rotate ($leafnode);
    $leafnode->Left->Type ($leafnode->Type);
    type ($leafnode->Right);
    $leafnode->Straighten_Recursive (int rand (4));
    print "    divide\n" if $DEBUG;
}

1;
