package Urb::Dom::Output;

use strict;
use warnings;

use Urb::Dom;
use SVG;

use vars qw /@ISA/;
@ISA = qw /Urb::Dom/;

=head1 NAME

Urb::Dom::Output - Various output file formats

=head1 SYNOPSIS

Graphic output for a domestic building subclass of L<Urb::Quad>

=head1 DESCRIPTION

=head1 USAGE

Create a new Urb::Dom::Output object:

  $dom = new Urb::Dom::Output;
  $child = new Urb::Quad::Output ($dom);

Dump the quad binary tree as a GraphViz .gv/.dot file:

  print STDOUT $quad->Dot_Tree;

=cut

sub Dot_Tree
{
    my $self = shift;
    my $label = shift;
    my $string = '';
    my $fill = '#666666';
    if ($self->Is_Outside) {$fill = '#ddddff'}
    elsif ($self->Type =~ /^b/i) {$fill = '#ff9999'}
    elsif ($self->Type =~ /^l/i) {$fill = '#99ff99'}
    elsif ($self->Type =~ /^k/i) {$fill = '#9999ff'}
    elsif ($self->Type =~ /^c/i) {$fill = '#ffff99'}
    elsif ($self->Type =~ /^t/i) {$fill = '#ff99ff'}

    $string .= '"'. $self->Level .'/'. $self->Id .' '. $self->Type .'" [color="'. $fill .'",style=filled];'. "\n" unless $self->Divided;
    return $string unless $self->Divided;
    $string .= "digraph G {\n" if (!defined $self->Parent and !defined $self->Below);
    $string .= "label=\"$label\"\n" if (defined $label and not defined $self->Parent);
    $string .= '"'. $self->Level .'/'. $self->Id .' '. $self->Type .'"->"'. $self->Left->Level .'/'. $self->Left->Id .' '. $self->Left->Type .'";'. "\n";
    $string .= '"'. $self->Level .'/'. $self->Id .' '. $self->Type .'"->"'. $self->Right->Level .'/'. $self->Right->Id .' '. $self->Right->Type .'";'. "\n";
    $string .= $self->Left->Dot_Tree;
    $string .= $self->Right->Dot_Tree;
    $string .= $self->Above->Dot_Tree if (!$self->Parent and defined $self->Above);
    $string .= '}' if (!defined $self->Parent and !defined $self->Below);
    return $string;
}

=pod

Dump the quad graph as an undirected GraphViz .gv/.dot file:

  print STDOUT $quad->Dot_Net;

=cut

sub Dot_Net
{
    my $self = shift;
    my $label = shift;
    my $string = "graph G {\n";
    $string .= "label=\"$label\"\n" if defined $label;

    for my $level ($self, $self->Levels_Above)
    {
        my $graph = $level->Graph;
        for my $vertex ($graph->vertices)
        {
            my $fill = '#666666';
            if ($vertex->Is_Outside) {$fill = '#ddddff'}
            elsif ($vertex->Type =~ /^b/i) {$fill = '#ff9999'}
            elsif ($vertex->Type =~ /^l/i) {$fill = '#99ff99'}
            elsif ($vertex->Type =~ /^k/i) {$fill = '#9999ff'}
            elsif ($vertex->Type =~ /^c/i) {$fill = '#ffff99'}
            elsif ($vertex->Type =~ /^t/i) {$fill = '#ff99ff'}
            $string .= '"'. $level->Level .'/'. $vertex->Id .' '. $vertex->Type .'" [color="'. $fill .'",style=filled];'. "\n";
        }
        for my $edge ($graph->edges)
        {
            $string .= '"'. $level->Level .'/'. $edge->[0]->Id .' '. $edge->[0]->Type .'"--"'. $level->Level .'/'. $edge->[1]->Id .' '. $edge->[1]->Type .'"' 
                .' [label="'. sprintf ("%.1f", $graph->get_edge_weight (@{$edge})) .'"];' ."\n";
        }
    }
    $string .= '}';
    return $string;
}

=pod

Create an L<SVG> object representing the current quad and children:

  $svg = $quad->SVG;

..or append the quad and children to an existing SVG object:

  $svg = $quad->SVG ($svg, $min, $max);

$min and $max are the crop window to retrieve from the quad.

=cut

sub SVG
{
    my $self = shift;
    my $scale = 20;
    my $svg = shift || new SVG (width => (($self->Max->[0] - $self->Min->[0] + $self->Levels_Above) * $scale),
                               height => (($self->Max->[1] - $self->Min->[1] + 1 + $self->Levels_Above) * $scale));
    my $min = shift || $self->Min;
    my $max = shift || $self->Max;
    my $height_c = $max->[1] - $min->[1];
    my $coor0 = Urb::Quad::_subtract_vector ($self->Coordinate (0), $min);
    my $coor1 = Urb::Quad::_subtract_vector ($self->Coordinate (1), $min);
    my $coor2 = Urb::Quad::_subtract_vector ($self->Coordinate (2), $min);
    my $coor3 = Urb::Quad::_subtract_vector ($self->Coordinate (3), $min);
    my $coor4 = Urb::Quad::_subtract_vector ($self->Centroid, $min);
    my $points = $svg->get_path (-type => 'path', -closed => 'true',
        x => [$coor0->[0] * $scale, $coor1->[0] * $scale, $coor2->[0] * $scale, $coor3->[0] * $scale],
        y => [($height_c - $coor0->[1]) * $scale, ($height_c - $coor1->[1]) * $scale,
              ($height_c - $coor2->[1]) * $scale, ($height_c - $coor3->[1]) * $scale]); 

    my $opacity = 0.5;
    my $width = 0.2 * $scale;
    my $fill = "rgb(0,0,64)";
    if ($self->Is_Outside)
    {
        $opacity = 0.3;
        $width = 0.0;
    }
    elsif ($self->Type =~ /^b/i)
    {
        $fill = "rgb(255,0,0)";
    }
    elsif ($self->Type =~ /^l/i)
    {
        $fill = "rgb(0,255,0)";
    }
    elsif ($self->Type =~ /^k/i)
    {
        $fill = "rgb(0,0,255)";
    }
    elsif ($self->Type =~ /^c/i)
    {
        $fill = "rgb(255,255,0)";
    }
    elsif ($self->Type =~ /^t/i)
    {
        $fill = "rgb(255,0,255)";
    }

    $svg->path (%$points,
                style => { 'fill-opacity' => $opacity,
                           'fill' => $fill,
                           'stroke-width' => $width,
                           'stroke' => "rgb(0,0,0)" })
        unless $self->Divided;

    $width = 0.2 * $scale;

    unless ($self->Parent)
    {
        for my $id (0 .. 3)
        {
            my $external = $self->Boundary_Id ($id);
            next unless (defined $self->Perimeter->{$external} and $self->Perimeter->{$external} =~ /^private$/i);
            my $a = Urb::Quad::_subtract_vector ($self->Coordinate (0 + $id), $min);
            my $b = Urb::Quad::_subtract_vector ($self->Coordinate (1 + $id), $min);
            my $p = $svg->get_path (-type => 'path',
                x => [$a->[0] * $scale, $b->[0] * $scale],
                y => [($height_c - $a->[1]) * $scale, ($height_c - $b->[1]) * $scale]); 

                $svg->path (%$p,
                style => { 'stroke-width' => 0.12 * $scale,
                           'stroke' => "rgb(0,0,0)" })
        }
        $svg->text (x => $coor4->[0] * $scale, y => (($max->[1] - $min->[1] + 0.6) * $scale), 'text-anchor' => 'middle' )->text (-type=>'span',
                style => { 'font-family' => 'serif',
                           'font-size' => $width * 2.5,
                           'fill' => "rgb(0,0,64)" })->cdata ($self->Hash) unless $self->Below;
        $svg = $self->Above->SVG ($svg, Urb::Quad::_subtract_vector ($min, [1,1]), Urb::Quad::_subtract_vector ($max, [1,1])) if defined $self->Above;
    }

    $svg->text (x => $coor4->[0] * $scale, y => ($height_c - $coor4->[1]) * $scale, 'text-anchor' => 'middle' )->text (-type=>'span',
                style => { 'font-family' => 'serif',
                           'font-size' => $width * 2,
                           'fill' => "rgb(0,0,64)" })->cdata (join ' ', $self->Level .'/'. $self->Id, $self->Type)
        unless $self->Divided;

    for my $child (@{$self->{child}})
    {
        $svg = $child->SVG ($svg, $min, $max);
    }

    return $svg;
}

=pod

Assemble the quad and children as a DXF fragment, if the quad is the root then
add a DXF header too:

  $string = $quad->DXF;

=cut

sub DXF
{
    my $self = shift;
    my $width = 0.1;
    my $thickness = 2.4;
    my $id = $self->Id;
    my $coor0 = $self->Coordinate (0);
    my $coor1 = $self->Coordinate (1);
    my $coor2 = $self->Coordinate (2);
    my $coor3 = $self->Coordinate (3);
    my $coor4 = $self->Centroid;

    my $string;

    $string .= qq|  0\nSECTION
  2\nHEADER
  9\n\$ACADVER
  1\nAC1009
  0\nENDSEC
  0\nSECTION
  2\nENTITIES
| unless (defined $self->Parent);

    $string .= qq| 0\nPOLYLINE
  8\n0
 66\n1
 10\n0.0
 20\n0.0
 30\n0.0
 39\n$thickness
 70\n1
 40\n$width
 41\n$width
  0\nVERTEX
  8\n0
 10\n$coor0->[0]
 20\n$coor0->[1]
 30\n0.0
  0\nVERTEX
  8\n0
 10\n$coor1->[0]
 20\n$coor1->[1]
 30\n0.0
  0\nVERTEX
  8\n0
 10\n$coor2->[0]
 20\n$coor2->[1]
 30\n0.0
  0\nVERTEX
  8\n0
 10\n$coor3->[0]
 20\n$coor3->[1]
 30\n0.0
  0\nSEQEND
  0\nTEXT
  8\n0
 10\n$coor4->[0]
 20\n$coor4->[1]
 30\n0.0
 40\n0.2
  1\n$id
| unless $self->Divided;

    for my $child (@{$self->{child}})
    {
        $string .= $child->DXF;
    }

    $string .= qq|  0\nENDSEC
  0\nEOF
| unless (defined $self->Parent);

    return $string;
}

=pod

Dump a quad binary tree in multiline format, optionally specify an indent string:

  die $quad->Pretty_Tree ('>');

=cut

# FIXME test needed
sub Pretty_Tree
{
    my $self = shift;
    my $indent = shift || '';
    my $string = $indent .'('. $self->Id .')';
    $string .= ' Type='. $self->Type if $self->Type;
    $string .= ' Rotation='. $self->Rotation;
    $string .= ' Area='. sprintf ("%.1f", $self->Area);
    $string .= ' Width='. sprintf ("%.1f", $self->Length_Narrowest);
    $string .= ' Covered' if $self->Is_Covered;
    $string .= ' Supported' if $self->Is_Supported;
    $string .= ' (LEAF)' unless $self->Divided;

    $string .= "\n";

    $string .= $self->Left->Pretty_Tree ($indent .'    ') if defined $self->Left;
    $string .= $self->Right->Pretty_Tree ($indent .'    ') if defined $self->Right;
    $string .= $self->Above->Pretty_Tree ($indent) if defined $self->Above and !$self->Parent;
    return $string;
}

1;
