package Urb::Misc::Ein;

# bruno@postle.net 2004-07-19
# 2004-10-04 fixed bug where $line was reread
# 2004-10-11 was choking on files with t-elements
# 2004-10-25 put ima values in a hash of their own, add SaveFile()
# 2004-11-15 fix reading of ein files without node fixity column. add MergeAll(). prototype struct in new()

use strict;
use warnings;

use Storable qw/ dclone /;
use Math::Trig qw/cartesian_to_cylindrical cylindrical_to_cartesian/;

=head1 NAME

Urb::Misc::Ein - TechNET EasyNT 'EIN' files

=head1 SYNOPSIS

Ein files represent FEA net nodes and links, this module provides methods for reading, manipulating and writing them.

=head1 DESCRIPTION

An EIN file consists of one or more 'groups' each with the same
structure.

Each group contains up to four blocks of data: IMA (including a
comment string), NODES, LINKS and T-ELEMENTS.

=head1 USAGE

Create a new Ein object like so:

   my $ein = new Urb::Misc::Ein;

This will be initialised with a single empty group.

=cut

sub new
{
    my $class = shift;
    $class = ref $class || $class;

    my %group = ( ima => { 1 => 0,  2 => 0,  3 => 0,  4 => 0,  5 => 0,
                           6 => 0,  7 => 0,  8 => 0,  9 => 0, 10 => 0,
                          11 => 0, 12 => 0, 13 => 0, 14 => 0, 15 => 0,
                          16 => 0, 17 => 0, 18 => 0, 19 => 0, 20 => 0},
                nodes => {},
                links => [],
           t_elements => [],
                  min => [],
                  max => [] );

    my $self = [{%group}];

    bless $self, $class;
    return $self;
}

=pod

Clone an object:

    my $ein_new = $ein->Clone;

=cut

sub Clone
{
    my $self = shift;
    dclone $self;
}

=pod

Parse an existing Ein file like so:

   $ein->LoadFile ('/path/to/STATIK.EIN');

=cut

sub LoadFile
{
    my $self = shift;
    my $file = shift;

    open FILE, "<". $file;
    my $state = "START";
    my $group = 0;

    my @lines = <FILE>;

    while (@lines)
    {
        if ($state eq "START")
        {
            $self->[$group]->{nodes} = {};
            $self->[$group]->{links} = [];
            $self->[$group]->{t_elements} = [];

            my $line = shift @lines;
            chomp $line;

            my @values = unpack ("A5A5A5A5A5A5A5A5A5A5", $line);

            for my $item (@values) { $item =~ s/ //g };

            $self->[$group]->{ima}->{1} = $values[0];
            $self->[$group]->{ima}->{2} = $values[1];
            $self->[$group]->{ima}->{3} = $values[2];
            $self->[$group]->{ima}->{4} = $values[3];
            $self->[$group]->{ima}->{5} = $values[4];
            $self->[$group]->{ima}->{6} = $values[5];
            $self->[$group]->{ima}->{7} = $values[6];
            $self->[$group]->{ima}->{8} = $values[7];
            $self->[$group]->{ima}->{9} = $values[8];
            $self->[$group]->{ima}->{10} = $values[9];

            $line = shift @lines;
            chomp $line;

            @values = unpack ("A5A5A5A5A5A5A5A5A5A5", $line);

            for my $item (@values) { $item =~ s/ //g };

            $self->[$group]->{ima}->{11} = $values[0];
            $self->[$group]->{ima}->{12} = $values[1];
            $self->[$group]->{ima}->{13} = $values[2];
            $self->[$group]->{ima}->{14} = $values[3];
            $self->[$group]->{ima}->{15} = $values[4];
            $self->[$group]->{ima}->{16} = $values[5];
            $self->[$group]->{ima}->{17} = $values[6];
            $self->[$group]->{ima}->{18} = $values[7];
            $self->[$group]->{ima}->{19} = $values[8];
            $self->[$group]->{ima}->{20} = $values[9];

            $line = shift @lines;
            chomp $line;

            $self->[$group]->{title} = $line;
            $state = "NODES";

            #print STDERR "found group $group ".
            #             "with ". ($self->[$group]->{ima}->{1} + $self->[$group]->{ima}->{2}) ." nodes, ".
            #             "$self->[$group]->{ima}->{4} links ".
            #            "and $self->[$group]->{ima}->{9} T-elements\n";
        }


        if ($state eq "NODES")
        {
            for my $node (1 .. ($self->[$group]->{ima}->{1} + $self->[$group]->{ima}->{2}))
            {
                my $line = shift @lines;
                chomp $line;

                my @values = unpack ("A12A15A15A15A4", $line); 

                for (@values) { $_ =~ s/ //g };
     
                my $item =  {xcoor => $values[1],
                             ycoor => $values[2],
                             zcoor => $values[3],
                            fixity => $values[4]};

                # mark as unfixed if unset, needs to correlated with
                $item->{fixity} = '111' unless $item->{fixity};
                $item->{fixity} = '000' if ($node > $self->[$group]->{ima}->{1});

                $self->[$group]->{nodes}->{$values[0]} = $item;
            }
            $state = "LINKS";
        }

        if ($state eq "LINKS")
        {
            for my $link (1 .. $self->[$group]->{ima}->{4})
            {
                my $line = shift @lines;
                chomp $line;

                my @values = unpack ("A10A10A10A10A10A10", $line);

                for (@values) { $_ =~ s/ //g };

                my $compressive = 'FALSE';
                   $compressive = 'TRUE' if ($link <= $self->[$group]->{ima}->{20});
                my $elastic = 'FALSE';
                   $elastic = 'TRUE' if ($link <= $self->[$group]->{ima}->{12});

                my $item =  {start => $values[0],
                               end => $values[1],
                            length => $values[2],
                     force_density => $values[3],
                         stiffness => $values[4],
                       input_field => $values[5],
                           elastic => $elastic,
                       compressive => $compressive};

                push @{$self->[$group]->{links}}, $item;
            }
            $state = "T-ELEMENTS";
        }

        if ($state eq "T-ELEMENTS")
        {
            for my $t_element (1 .. $self->[$group]->{ima}->{9})
            {
                my $line = shift @lines;
                chomp $line;

                my @values = unpack ("A10A10A10", $line);

                for (@values) { $_ =~ s/ //g };
     
                my $item =  {start => $values[0],
                               end => $values[1],
                          position => $values[2]};

                push @{$self->[$group]->{t_elements}}, $item;
            }
            $state = "START";
            $group++;
        }
    }

    close FILE;
}

=pod

After manipulating data, the IMA tables for each group can be
updated with the Update command:

   $ein->Update;

=cut

sub Update
{
    my $self = shift;

    for my $group (@$self)
    {
        my $nodes_not_fully_fixed = 0;
        for my $nodeid (keys %{$group->{nodes}})
        {
            my $node = $group->{nodes}->{$nodeid};
            next if ($node->{fixity} eq "000");
            $nodes_not_fully_fixed++;
        }

        my $nodes_fully_fixed = 0;
        for my $nodeid (keys %{$group->{nodes}})
        {
            my $node = $group->{nodes}->{$nodeid};
            next unless ($node->{fixity} eq "000");
            $nodes_fully_fixed++;
        }

        my $min = $group->{min};
        my $max = $group->{max};

        for my $nodeid (keys %{$group->{nodes}})
        {
            my $node = $group->{nodes}->{$nodeid};
            $min->[0] = $node->{xcoor}
                if (not defined $min->[0] or $node->{xcoor} < $min->[0]);
            $min->[1] = $node->{ycoor}
                if (not defined $min->[1] or $node->{ycoor} < $min->[1]);
            $min->[2] = $node->{zcoor}
                if (not defined $min->[2] or $node->{zcoor} < $min->[2]);
            $max->[0] = $node->{xcoor}
                if (not defined $max->[0] or $node->{xcoor} > $max->[0]);
            $max->[1] = $node->{ycoor}
                if (not defined $max->[1] or $node->{ycoor} > $max->[1]);
            $max->[2] = $node->{zcoor}
                if (not defined $max->[2] or $node->{zcoor} > $max->[2]);
        }

        my $links = @{$group->{links}};
        my $t_elements = @{$group->{t_elements}};

        my $links_compressive = 0;
        for my $link (@{$group->{links}})
        {
            next unless ($link->{compressive} eq 'TRUE');
            $links_compressive++;
        }

        my $links_elastic_fixed_length = 0;
        for my $link (@{$group->{links}})
        {
            next unless ($link->{elastic} eq 'TRUE');
            $links_elastic_fixed_length++;
        }

=pod

Most of the IMA table isn't used.  However, these slots are used to
indicate the number of each type of element:

   1 - nodes that don't have full fixity
   2 - fully fixed nodes
   4 - links
   9 - t-elements
  12 - elastic links with fixed lengths
  20 - compressive links.

There is no space for more than 20 IMA slots.

=cut

        my $ima = $group->{ima};
        $ima->{1} = $nodes_not_fully_fixed;
        $ima->{2} = $nodes_fully_fixed;
        $ima->{4} = $links;
        $ima->{9} = $t_elements;
        $ima->{12} = $links_elastic_fixed_length;
        $ima->{20} = $links_compressive;
    }
}


our $ima;
format IMA =
@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>
$ima->{1}, $ima->{2}, $ima->{3}, $ima->{4}, $ima->{5}, $ima->{6}, $ima->{7}, $ima->{8}, $ima->{9}, $ima->{10}
@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>@>>>>
$ima->{11}, $ima->{12}, $ima->{13}, $ima->{14}, $ima->{15}, $ima->{16}, $ima->{17}, $ima->{18}, $ima->{19}, $ima->{20}
perl-Easy:     this is a comment
.

our $node;
format NODE =
@>>>>>>>>>>>@#######.######@#######.######@#######.######@>>>
$node->{id},$node->{xcoor},$node->{ycoor},$node->{zcoor},$node->{fixity}
.

our $link;
format LINK =
@>>>>>>>>>@>>>>>>>>>@##.######@####.####@######.##@>>>>>>>>>
$link->{start},$link->{end},$link->{length},$link->{force_density},$link->{stiffness},$link->{input_field}
.

our $t_element;
format T_ELEMENT =
@>>>>>>>>>@>>>>>>>>>@###.#####
$t_element->{start},$t_element->{end},$t_element->{position}
.

=pod

Write an Ein file like so:

   $ein->SaveFile ('/path/to/output.ein');

=cut

sub SaveFile
{
    my $self = shift;
    my $file = shift;

    $self->Update;

    open FILE, ">". $file;
    binmode FILE, ":crlf";
    select FILE;

    for my $group (@$self)
    {
        $ima = $group->{ima};
        $~ = 'IMA'; write;
        for my $nodeid (sort {$a <=> $b} keys %{$group->{nodes}})
        {
            $node = $group->{nodes}->{$nodeid};
            $node->{id} = $nodeid;
            next if ($node->{fixity} eq "000");
            $~ = 'NODE'; write;
        }

        for my $nodeid (sort {$a <=> $b} keys %{$group->{nodes}})
        {
            $node = $group->{nodes}->{$nodeid};
            $node->{id} = $nodeid;
            next unless ($node->{fixity} eq "000");
            $~ = 'NODE'; write;
        }

        warn "EasyNT has a limit of 10000 nodes"
            if (scalar (keys %{$group->{nodes}}) > 10000);

        for my $linkey (@{$group->{links}})
        {
            warn "Stiffness > 10000000 unsupported by EIN"
                if ($linkey->{stiffness} >= 10000000);
        }
        for my $linkey (@{$group->{links}})
        {
            $link = $linkey;
            next unless ($link->{compressive} eq "TRUE");
            next unless ($link->{elastic} eq "TRUE");
            $~ = 'LINK'; write;
        }
        for my $linkey (@{$group->{links}})
        {
            $link = $linkey;
            next unless ($link->{compressive} eq "FALSE");
            next unless ($link->{elastic} eq "TRUE");
            $~ = 'LINK'; write;
        }
        for my $linkey (@{$group->{links}})
        {
            $link = $linkey;
            next unless ($link->{compressive} eq "FALSE");
            next unless ($link->{elastic} eq "FALSE");
            $~ = 'LINK'; write;
        }

        for my $t_elementey (@{$group->{t_elements}})
        {
            $t_element = $t_elementey;
            $~ = 'T_ELEMENT'; write;
        }

        warn "EasyNT has a limit of 2000 T-elements"
            if (scalar @{$group->{t_elements}} > 2000);
    }

    close FILE;
}

=head1 UTILITIES

Join all groups into a single group:

   $ein = $ein->MergeAll;

Nodes with identical IDs will be merged, duplicate links will be
retained.

=cut

sub MergeAll
{
    my $self = shift;
    my $result = new Urb::Misc::Ein;

    $self->Update;

    for my $group (@$self)
    {
        for my $nodeid (keys %{$group->{nodes}})
        {
            $result->[0]->{nodes}->{$nodeid} = $group->{nodes}->{$nodeid};
        }

        for my $link (@{$group->{links}})
        {
            push @{$result->[0]->{links}}, $link;
        }

        for my $t_element (@{$group->{t_elements}})
        {
            # this is broken (bruno 2009-01)
            push @{$result->[0]->{t_elements}}, $t_element;
        }
    }
    $result->Update;
    return $result;
}

sub Difference
{
    my $self = shift;
    my $deflected = shift || return 0;
    my $result = new Urb::Misc::Ein;

    for my $nodeid (keys %{$self->[0]->{nodes}})
    {
        my $node_s = $self->[0]->{nodes}->{$nodeid};
        my $node_d = $deflected->[0]->{nodes}->{$nodeid};
        #my $distance = distance ([$node_s->{xcoor},$node_s->{ycoor},$node_s->{zcoor}],
        #                         [$node_d->{xcoor},$node_d->{ycoor},$node_d->{zcoor}]); 
        #
        #$result->[0]->{nodes}->{$nodeid} = {xcoor => $node_s->{xcoor},
        #                                    ycoor => $node_s->{ycoor},
        #                                    zcoor => $distance,
        #                                   fixity => $node_s->{fixity}};
        $result->[0]->{nodes}->{$nodeid} = {xcoor => ($node_d->{xcoor}),
                                            ycoor => ($node_d->{ycoor}),
                                            zcoor => ($node_d->{zcoor} - $node_s->{zcoor}),
                                           fixity => $node_s->{fixity}};
    }

    for my $link (@{$self->[0]->{links}})
    {
        push @{$result->[0]->{links}}, $link;
    }

    for my $t_element (@{$self->[0]->{t_elements}})
    {
        push @{$result->[0]->{t_elements}}, $t_element;
    }

    $result->Update;
    return $result;
}

sub distance
{
    my ($a, $b) = @_;
    sqrt (square ($a->[0] - $b->[0])
        + square ($a->[1] - $b->[1])
        + square ($a->[2] - $b->[2]));
}

sub square
{
    my $number = shift;
    $number * $number;
}

=pod

Rotate a mesh around z axis

    $ein->RotateZ ($group, 3.141);

=cut

sub RotateZ
{
    my $self = shift;
    my $group = shift;
    my $angle = shift;
    for my $node_id (keys %{$self->[$group]->{nodes}})
    {
        my $node = $self->[$group]->{nodes}->{$node_id};
        my ($rho, $theta, $z) = cartesian_to_cylindrical ($node->{xcoor}, $node->{ycoor}, $node->{zcoor});
        $theta += $angle;
        ($node->{xcoor}, $node->{ycoor}, $node->{zcoor}) = cylindrical_to_cartesian ($rho, $theta, $z);
        
    }
}

=pod

Access the coordinates of any node like so:

    my $group = 0;
    my $vector = $ein->Node ($group, 33212);

(Note that group numbering starts at zero)

=cut

sub Node
{
    my $self = shift;
    my $group = shift;
    my $id = shift;

    my $node = $self->[$group]->{nodes}->{$id} || return 0;
    return [$node->{xcoor}, $node->{ycoor}, $node->{zcoor}];
}

=pod

Access the connections of any node like so:

    my $group = 0;
    my $node_ids = $ein->Connections ($group, 33212);

=cut

sub Connections
{
    my $self = shift;
    my $group = shift;
    my $id = shift;

    my $nodes = [];

    for my $link (@{$self->[$group]->{links}})
    {
        push (@{$nodes}, $link->{end}) if ($link->{start} == $id);
        push (@{$nodes}, $link->{start}) if ($link->{end} == $id);
    }
    return $nodes;
}

=pod

Access the next unused node number after given number

    my $node_id = $ein->NextNode (9000);

=cut

sub NextNode
{
    my $self = shift;
    my $index = shift;

    while (1)
    {
        my $used = 'UNKNOWN';
        for my $group (@{$self})
        {
            $used = 'TRUE' if defined $group->{nodes}->{$index};
        }
        return $index unless $used eq 'TRUE';
        $index++;
    }
}

=pod

Access or set the length of a link:

    my $length = $ein->Length ($group, 123);
    $ein->Length ($group, 123, 4.56);

Note: this is a link attribute, not necessarily the distance between the ends

=cut

sub Length
{
    my $self = shift;
    my $group = shift;
    my $link_id = shift;
    my $length = shift;

    my $link = $self->[$group]->{links}->[$link_id];
    $link->{length} = $length if (defined $length);

    return $link->{length};
}

=pod

Access the strain in a link, 0.0 is no strain, positive is extension, negative is compression

  my $strain = $self->Strain ($group_id, 123);

=cut

sub Strain
{
    my $self = shift;
    my $group_id = shift;
    my $link_id = shift;
    my $length = $self->Length ($group_id, $link_id);
    my $distance = $self->Distance ($group_id,
                                    $self->[$group_id]->{links}->[$link_id]->{start},
                                    $self->[$group_id]->{links}->[$link_id]->{end});
    return ($distance/$length) -1;
}

=pod

Access the load in a link, positive values are tension:

    my $P = $ein->Load ($group, 123);

(calculated by multiplying length by load density)

=cut

sub Load
{
    my $self = shift;
    my $group = shift;
    my $link_id = shift;
    my $link = $self->[$group]->{links}->[$link_id];
    return $link->{length} * $link->{force_density};
}

=pod

Get a list of links connecting any two node IDs

    my $links = $ein->LinksMatching ($group, 123, 456);

=cut

# TODO tests

sub LinksMatching
{
    my $self = shift;
    my $group = shift;
    my ($a, $b) = @_;

    my @result;
    my $id_link = 0;
    for my $link (@{$self->[$group]->{links}})
    {
        push @result, $id_link if ($link->{start} == $a and $link->{end} == $b); 
        push @result, $id_link if ($link->{start} == $b and $link->{end} == $a); 
        $id_link++;
    }
    return [@result];
}

=pod

Query the geometrical distance between any two nodes:

    my $distance = $ein->Distance ($group, 234, 567);

=cut

sub Distance
{
    my $self = shift;
    my $group = shift;
    my ($start, $end) = @_;    
    
    distance ($self->Node ($group, $start), $self->Node ($group, $end));
}

=pod

Query the nodes of any link like so:

    my @nodes = $ein->Nodes ($group, $link_index);
or:
    my $nodes = $ein->Nodes ($group, $link_index);

=cut

sub Nodes
{
    my $self = shift;
    my $group = shift;
    my $link_index = shift;
    my $link = $self->[$group]->{links}->[$link_index];

    return wantarray ? ($link->{start}, $link->{end}) : [$link->{start}, $link->{end}];
}

=pod

Query the number of links in a group

    my $links = $ein->Links ($group);

=cut

sub Links
{
    my $self = shift;
    my $group = shift;

    scalar (@{$self->[$group]->{links}});
}


=pod

Access the minimum and maximum coordinates of any group:

    my $min = $ein->Min ($group);
    my $max = $ein->Max ($group);

If the group is unspecified, all groups are considered:

    my $MAX = $ein->Max;

=cut

sub Min
{
    my $self = shift;
    my $group_id = shift;
    return $self->[$group_id]->{min} if defined $group_id;

    my $min = [@{$self->[0]->{min}}];
    for my $group (@$self)
    {
        $min->[0] = $group->{min}->[0] if $group->{min}->[0] < $min->[0];
        $min->[1] = $group->{min}->[1] if $group->{min}->[1] < $min->[1];
        $min->[2] = $group->{min}->[2] if $group->{min}->[2] < $min->[2];
    }
    return $min;
}

sub Max
{
    my $self = shift;
    my $group_id = shift;

    return $self->[$group_id]->{max} if defined $group_id;

    my $max = [@{$self->[0]->{max}}];
    for my $group (@$self)
    {
        $max->[0] = $group->{max}->[0] if $group->{max}->[0] > $max->[0];
        $max->[1] = $group->{max}->[1] if $group->{max}->[1] > $max->[1];
        $max->[2] = $group->{max}->[2] if $group->{max}->[2] > $max->[2];
    }
    return $max;
}


=pod

=head1 COPYRIGHT

Copyright (c) 2004 Bruno Postle <bruno@postle.net>.  All Rights Reserved.
This module is Free Software. It may be used, redistributed and/or
modified under the same terms as Perl itself.

=cut

1;

