package Urb::Polygon;

use strict;
use warnings;
use 5.010;

use Math::Trig;
use Math::Polygon;

our $PI = atan2 (0,-1);

use vars qw /@ISA/;
@ISA = qw /Math::Polygon/;

=head1 NAME

Urb::Polygon - Class for maintaining polygon data

=head1 SYNOPSIS

Some dodgy new Math::Polygon methods

=head1 DESCRIPTION

This module contains some extra methods for L<Math::Polygon>, these methods are
not quite sane enough to be added to L<Math::Polygon> proper.

=head1 USAGE

Get list of line gradients in use in this polygon, may contain duplicates

  @scalars = $polygon->_gradients;

=cut

sub _gradients
{
    my $self = shift;
    return map {sprintf ("%.8f", $_->{a})} $self->_lines;
}

=pod

Get list of lines in use in this polygon in a,b format. may contain duplicates

  @lines = $polygon->_lines;

=cut

sub _lines
{
    my $self = shift;
    my @points = $self->points;
    my @lines;
    for my $id (0 .. scalar (@points) -2)
    {
        push @lines, _points_2line ($points[$id], $points[$id+1]);
    }
    return @lines;
}

=pod

See if all corners of this polygon are on edges of or inside another

  $polygon->_is_contained_by ($bigger_polygon);

=cut

sub _is_contained_by
{
    my $self = shift;
    my $bigger = shift;
    my @points = $self->points;
    pop @points;
    for my $point (@points)
    {
        return 0 unless ($bigger->contains ($point) || $bigger->_is_on_polygon ($point));
    }
    return 1;
}

=pod

Do two polygons overlap? crude test

  $polygon->_overlaps ($other_polygon);

=cut

sub _overlaps
{
    my $self = shift;
    my $other = shift;

    my $other_divided = $other->_divide_polygon (20);

    my @points = $other_divided->points;
    pop @points;
    for my $point (@points)
    {
       return 1 if ($self->contains ($point) && !$self->_is_on_polygon ($point));
    }

    return 0;
}

=pod

Is a point on the edge of this polygon?

   $polygon->_is_on_polygon ([2.3,4.5]);

=cut

sub _is_on_polygon
{
    my $self = shift;
    my $point = shift;
    my @points = $self->points;
    pop @points;
    for my $id (0 .. scalar (@points) -1)
    {
        my $length = _distance_2d ($points[$id-1], $points[$id]);
        my $length_a = _distance_2d ($points[$id-1], $point);
        my $length_b = _distance_2d ($points[$id], $point);
        return 1 if abs ($length - $length_a - $length_b) < 0.000001;
    }
    return 0;
}

=pod

Returns a new polygon with all edges subdivided

   $polygon->_divide_polygon (4);

=cut

sub _divide_polygon
{
    my $self = shift;
    my $n = shift || 2;
    my @points = $self->points;
    my @out = ($points[0]);
    for my $id (0 .. scalar (@points) -2)
    {
        my $step = _scale_2d (1/$n, _subtract_2d ($points[$id+1], $points[$id]));
        for my $inc (1 .. $n)
        {
            my $vector = _scale_2d ($inc, $step);
            my $point = _add_2d ($points[$id], $vector);
            push @out, $point;
        }
    }

    my $out = new Urb::Polygon (points => [@out]);
    return $out;
}

=pod

Deconstruct polygon into overlapping quads, supply list of 'holes' if
necessary, holes don't need to be entirely within outside polygon

   my @quads = $polygon->_to_quads ($hole_polygon);

=cut

sub _to_quads
{
    my $polygon = shift;
    my @inners = @_;

    # make a list of the line-equations of all perimeter segments
    my @lines_all = map {$_->_lines} ($polygon, @inners);

    # iterate through all combinations of four lines
    my $shapes;
    for my $A (0 .. scalar @lines_all -1)
    {
      for my $B ($A +1 .. scalar @lines_all -1)
      {
        for my $C ($B +1 .. scalar @lines_all -1)
        {
          for my $D ($C +1 .. scalar @lines_all -1)
          {
            my $line_a = $lines_all[$A];
            my $line_b = $lines_all[$B];
            my $line_c = $lines_all[$C];
            my $line_d = $lines_all[$D];

            # get all possible intersections of these four lines
            my @corners;
            push @corners, _line_intersection ($line_a,$line_b) unless $line_a->{a} == $line_b->{a};
            push @corners, _line_intersection ($line_a,$line_c) unless $line_a->{a} == $line_c->{a};
            push @corners, _line_intersection ($line_a,$line_d) unless $line_a->{a} == $line_d->{a};
            push @corners, _line_intersection ($line_c,$line_b) unless $line_c->{a} == $line_b->{a};
            push @corners, _line_intersection ($line_c,$line_d) unless $line_c->{a} == $line_d->{a};
            push @corners, _line_intersection ($line_b,$line_d) unless $line_b->{a} == $line_d->{a};

            # discard any points outside the perimeter
            my @shape;
            for my $corner (@corners)
            {
                next unless ($polygon->contains ($corner) || $polygon->_is_on_polygon ($corner));
                push @shape, $corner;
            }
            # skip if we can't make a quadrilateral from these points
            next unless scalar @shape == 4;
     
            # arrange the four points in a non-crossing polygon
            my $shape_a = new Urb::Polygon
                (points => [$shape[0], $shape[1], $shape[2], $shape[3], $shape[0]]);
            my $shape_b = new Urb::Polygon
                (points => [$shape[0], $shape[1], $shape[3], $shape[2], $shape[0]]);
            my $shape_c = new Urb::Polygon
                (points => [$shape[0], $shape[2], $shape[1], $shape[3], $shape[0]]);

            my @biggest_first = sort {$b->area <=> $a->area} ($shape_a, $shape_b, $shape_c);        
            my $shape = shift @biggest_first;

            # normalise this shape and save for later
            $shape->counterClockwise;
            $shape = $shape->startMinXY;
            push @{$shapes}, $shape;
          }
        }
      }
    }

    my $shapes_clean = [];

    my $lookup;

    my @polygon_gradients = ($polygon->_gradients, map {$_->_gradients} @inners);

    # loop to remove some obviously bogus polygons
    for my $shape (@{$shapes})
    {
        my $okayness = 1;

        # characterise this polygon and skip if already seen
        my $key = join '-', (map {join ':', map {sprintf ("%.3f", $_)} @{$_}} $shape->points);
        next if defined $lookup->{$key};
        $lookup->{$key} = 1;

        # all edges must be parallel to a perimeter edge
        for my $gradient ($shape->_gradients)
        {
            $okayness = 0 unless grep /^$gradient$/, @polygon_gradients;
        }
        next unless $okayness;

        # shape can't overlap an inner perimeter
        for my $inner (@inners)
        {
            my $inner_normalised = $inner->startMinXY;
            $inner_normalised->counterClockwise;
            $okayness = 0 if $shape->_overlaps ($inner_normalised);
            $okayness = 0 if $inner_normalised->_overlaps ($shape);
            $okayness = 0 if $shape->same ($inner_normalised);
        }
        # shape can't contain outer perimeter
        $okayness = 0 if $shape->_overlaps ($polygon);
        next unless $okayness;

        # shape shouldn't cover an outside area
        my $centre = _scale_2d (0.5, _add_2d (($shape->points)[0], ($shape->points)[2]));
        $okayness = 0 unless $polygon->contains ($centre);
        next unless $okayness;

        push @{$shapes_clean}, $shape if $okayness;
    }

    @{$shapes} = sort {$b->area <=> $a->area} @{$shapes_clean};

    $shapes_clean = [];

    # remove any shapes that are contained within larger shapes
    for my $shape (@{$shapes})
    {
        my $okayness = 1;
        for my $bigger (@{$shapes_clean})
        {
            $okayness = 0 if $shape->_is_contained_by ($bigger);
        }
        push @{$shapes_clean}, $shape if $okayness;
    }

    return @{$shapes_clean};
}

# vector math

sub _is_between_2d
{
    my $point = shift;
    my @points = @_;
    my $length = _distance_2d ($points[0], $points[1]);
    my $length_a = _distance_2d ($points[0], $point);
    my $length_b = _distance_2d ($points[1], $point);
    return 1 if abs ($length - $length_a - $length_b) < 0.000001;
    return 0;
}

sub _distance_2d
{
    my ($a, $b) = @_;
    sqrt (($a->[0] - $b->[0])**2 + ($a->[1] - $b->[1])**2);
}

sub _angle_2d
{
    my ($a, $b) = @_;
    return ($PI/2) if (($a->[0] == $b->[0]) and ($a->[1] < $b->[1]));
    return ($PI/-2) if (($a->[0] == $b->[0]) and ($a->[1] > $b->[1]));
    my $radians = atan (($b->[1] - $a->[1]) / ($b->[0] - $a->[0]));
    $radians += $PI if $b->[0] < $a->[0];

    return $radians;
}

sub _subtract_2d
{
    my ($a, $b) = @_;
    return [$a->[0] - $b->[0], $a->[1] - $b->[1]];
}

sub _add_2d
{
    my ($a, $b) = @_;
    return [$b->[0] + $a->[0], $b->[1] + $a->[1]];
}

sub _scale_2d
{
    my ($factor, $a) = @_;
    ($factor, $a) = ($a, $factor) if ref $factor;
    return [$a->[0] * $factor, $a->[1] * $factor];
}

sub _normalise_2d
{
    my $a = shift;
    my $length = _distance_2d ([0,0], $a);
    $length = 0.00000000001 unless $length;
    return [$a->[0] / $length, $a->[1] / $length];
}

sub _det
{
    my ($a, $b) = @_;
    return ($a->[0] * $b->[1]) - ($a->[1] * $b->[0]);
}

# line math

sub _points_2line
{
    my ($coor_0, $coor_1) = @_;
    my $vector = _subtract_2d ($coor_1, $coor_0);
    $vector->[0] = 0.00000000001 unless $vector->[0];
    my $a = $vector->[1] / $vector->[0];
    my $b = $coor_0->[1] - ($coor_0->[0] * $a);
    {a => $a, b => $b};
}

sub _perpendicular_line
{
    my ($line, $coor) = @_;
    $line->{a} = 0.00000000001 unless $line->{a};
    my $a = -1/$line->{a};
    my $b = $coor->[1] - ($coor->[0] * $a);
    {a => $a, b => $b};
}

sub _line_intersection
{
    my ($line_0, $line_1) = @_;
    return undef if ($line_0->{a} == $line_1->{a});
    my $x = ($line_1->{b} - $line_0->{b}) / ($line_0->{a} - $line_1->{a});
    my $y = ($line_0->{a} * $x) + $line_0->{b};
    [$x, $y];
}

sub _perpendicular_distance
{
    my ($line, $coor) = @_;
    my $perpendicular_line = _perpendicular_line ($line, $coor);
    my $intersection = _line_intersection ($line, $perpendicular_line);
    _distance_2d ($coor, $intersection);
}

1;
