package Urb::Route;
use strict;
use warnings;

use Urb::Misc::Ein;
use Graph::Undirected;
use List::Util;
use Math::Trig;
use Math::Polygon;

use vars qw /@ISA/;
@ISA = qw /Graph::Undirected/;

our $PI = Math::Trig::pi;
our $PI2 = 2 * $PI;

sub new
{
    my $class = shift;
    $class = ref $class || $class;

    my $self = new Graph::Undirected;

    bless $self, $class;
    return $self;
}

# FIXME Ein files are completely inappropriate for this purpose

sub Read_Ein
{
    my $self = shift;
    $self->delete_vertices ($self->vertices);
    my $path_ein = shift || return 0;

    my $ein = new Urb::Misc::Ein;
    $ein->LoadFile ($path_ein);

    for my $id_node (keys %{$ein->[0]->{nodes}})
    {
        $self->set_vertex_attribute ($id_node, 'coor', $ein->Node (0, $id_node));
    }

    for my $link (@{$ein->[0]->{links}})
    {
        $self->add_edge ($link->{start}, $link->{end});
    }
    return 1;
}

sub Write_Ein
{
    my $self = shift;
    my $path_ein = shift || return 0;

    my $ein = new Urb::Misc::Ein;

    for my $vertex ($self->vertices)
    {
        my ($x, $y, $z) = @{$self->get_vertex_attribute ($vertex, 'coor')};
        $ein->[0]->{nodes}->{$vertex} =
                {xcoor => $x,
                 ycoor => $y,
                 zcoor => $z,
                 fixity => '000'};
    }

    my $link_id = 9000;

    for my $edge ($self->edges)
    {
        my $node0 = $edge->[0];
        my $node1 = $edge->[1];

                my $link =  {start => $node0,
                               end => $node1,
                            length => 1,
                     force_density => ($self->get_edge_attribute ($node0, $node1, 'heat') || 1),
                         stiffness => 5000,
                       input_field => $link_id,
                           elastic => 'FALSE',
                       compressive => 'FALSE'};
        push @{$ein->[0]->{links}}, $link;
        $link_id ++;
    }

    $ein->Update;
    $ein->SaveFile ($path_ein);
}

# Travel between all possible destinations using shortest paths and assign a
# 'heat' quantity for every edge depending on resulting usage

sub Calc_Heat
{
    my $self = shift;
    $self->SPT_Dijkstra_clear_cache;
    my @vertices = $self->vertices;

    my @edges = $self->edges;
    $self->set_edge_attribute (@{$_}, 'heat', 0.0) for @edges;

    for my $id_a (0 .. scalar @vertices -2)
    {
        for my $id_b ($id_a +1 .. scalar @vertices -1)
        {
            my $relevance = 0.0001 * $self->get_vertex_attribute ($vertices[$id_a], 'length') * $self->get_vertex_attribute ($vertices[$id_b], 'length');
            my @path = $self->SP_Dijkstra ($vertices[$id_a], $vertices[$id_b]);
            for my $id_path (0 .. scalar @path -2)
            {
                my $heat = $self->get_edge_attribute ($path[$id_path], $path[$id_path +1], 'heat');
                $heat += $relevance;
                $self->set_edge_attribute ($path[$id_path], $path[$id_path +1], 'heat', $heat);
            }
        }
    }
}

# Dijkstra shortest path needs weights assigned to each edge.  nodes have
# cartesian coordinates, so calculate weight from edge length.

sub Calc_Weights
{
    my $self = shift;

    for my $edge ($self->edges)
    {
        my $length = $self->Get_Distance (@{$edge});

        $self->set_edge_attribute (@{$edge}, 'length', $length);
        $self->set_edge_weight (@{$edge}, $length);
    }

    for my $vertex ($self->vertices)
    {
        my $length = 0.0;
        for my $neighbour ($self->neighbours ($vertex))
        {
            $length += $self->get_edge_attribute ($vertex, $neighbour, 'length');
        }
        $self->set_vertex_attribute ($vertex, 'length', $length);
    }
}

# get distance between two vertices assuming they have 'coor' attributes

sub Get_Distance
{
    my $self = shift;
    my ($a, $b) = @_;
    _distance ($self->get_vertex_attribute ($a, 'coor'), $self->get_vertex_attribute ($b, 'coor'));
}

=pod

50 T JUNCTIONS

Lay out the road system so that any two roads which
meet at grade, meet at three-way T junctions as near
as 90 degrees as possible.  Avoid four-way intersections and
crossing movements.

=cut

sub Slide
{
    my $self = shift;
    for my $vertex (List::Util::shuffle ($self->vertices))
    {
        my @neighbours = $self->neighbours ($vertex);
        next unless scalar @neighbours == 3;
        my $weight_a = $self->get_edge_attribute ($vertex, $neighbours[0], 'heat');
        my $weight_b = $self->get_edge_attribute ($vertex, $neighbours[1], 'heat');
        my $weight_c = $self->get_edge_attribute ($vertex, $neighbours[2], 'heat');

        my $coor_a = $self->get_vertex_attribute ($neighbours[0], 'coor');
        my $coor_b = $self->get_vertex_attribute ($neighbours[1], 'coor');
        my $coor_c = $self->get_vertex_attribute ($neighbours[2], 'coor');
        my $coor = $self->get_vertex_attribute ($vertex, 'coor');

        my $dist_a = $self->get_edge_attribute ($vertex, $neighbours[0], 'length');
        my $dist_b = $self->get_edge_attribute ($vertex, $neighbours[1], 'length');
        my $dist_c = $self->get_edge_attribute ($vertex, $neighbours[2], 'length');

        if (($weight_a < $weight_b) and ($weight_a < $weight_c))
        {
            $weight_a = 0.2;
            $weight_b = 1.0;
            $weight_c = 1.0;
        }
        elsif (($weight_b < $weight_a) and ($weight_b < $weight_c))
        {
            $weight_a = 1.0;
            $weight_b = 0.2;
            $weight_c = 1.0;
        }
        else
        {
            $weight_a = 1.0;
            $weight_b = 1.0;
            $weight_c = 0.2;
        }

        $weight_a = 0.1 if $dist_a < 40.0;
        $weight_b = 0.1 if $dist_b < 40.0;
        $weight_c = 0.1 if $dist_c < 40.0;

        my $weight_total = $weight_a + $weight_b + $weight_c;
        $weight_a /= $weight_total;
        $weight_b /= $weight_total;
        $weight_c /= $weight_total;

        $coor_a = _scale_vector ($coor_a, $weight_a);
        $coor_b = _scale_vector ($coor_b, $weight_c);
        $coor_c = _scale_vector ($coor_c, $weight_b);
        my $coor_new = _add_vectors ($coor_a, $coor_b, $coor_c);

        $coor = _add_vectors (_scale_vector ($coor, 0.8), _scale_vector ($coor_new, 0.2));

        $self->set_vertex_attribute ($vertex, 'coor', $coor);
    }
}

# Find a high crinkliness polygon and remove it by merging it with an adjacent polygon by
# removing a link.  Link to be removed is link with least 'heat'.

sub Reap
{
    my $self = shift;
    my $crinkliness = shift || 0.08;
    my $loops = $self->Loops;
    for my $loop (@{$loops})
    {
        my $polygon = $self->Polygon ($loop);
        next unless $polygon->crinkliness > $crinkliness;
        my @todie;
        my $heat = 999999999999999;
        for my $i (0 .. scalar @{$loop} -2)
        {
            my $heat_this = $self->get_edge_attribute ($loop->[$i], $loop->[$i+1], 'heat');
            next unless $heat_this < $heat;
            next unless scalar $self->neighbours ($loop->[$i]) == 3;
            next unless scalar $self->neighbours ($loop->[$i+1]) == 3;
            @todie = ($loop->[$i], $loop->[$i+1]);
            $heat = $heat_this;
        }
        next unless scalar @todie;
        next unless $self->Delete_Edge (@todie);
        print STDERR "deleting polygon with crinkliness ". $polygon->crinkliness . "(> $crinkliness)\n";
        last;
    }
}

sub Crooked
{
    my $self = shift;
    my $angle_threshold = shift || ($PI / 4);
    my $loops = $self->Loops;
    for my $loop (@{$loops})
    {
        my $polygon = $self->Polygon ($loop);
        my @todie = ($loop->[0], $loop->[1]);
        my $angle = $self->Angle ($loop->[-2], $loop->[0], $loop->[1]); # -1 should have same location as 0
        for my $i (0 .. scalar @{$loop} -3)
        {
            my $angle_this = $self->Angle ($loop->[$i], $loop->[$i+1], $loop->[$i+2]);
            next unless $angle_this < $angle;
            next unless scalar $self->neighbours ($loop->[$i]) == 3;
            next unless scalar $self->neighbours ($loop->[$i+1]) == 3;
            @todie = ($loop->[$i], $loop->[$i+1]);
            $angle = $angle_this;
        }
        next unless $angle < $angle_threshold;
        next unless $self->Delete_Edge (@todie);
        print STDERR "deleting edge with angle ". $angle . "\n";
        last;
    }
}

# deletes two linked vertices (hence five edges if all vertices have three edges),
# replaces with two edges using remaining vertices

sub Delete_Edge
{
    my $self = shift;
    my @todie = @_;
    my @neighbours0 = $self->neighbours ($todie[0]);
    my $other = $todie[1];
    return 0 unless scalar @neighbours0 == 3;
    @neighbours0 = grep !/^$other$/, @neighbours0;
    return 0 unless scalar @neighbours0 == 2;
    
    my @neighbours1 = $self->neighbours ($todie[1]);
    $other = $todie[0];
    return 0 unless scalar @neighbours1 == 3;
    @neighbours1 = grep !/^$other$/, @neighbours1;
    return 0 unless scalar @neighbours1 == 2;

    $self->add_edge (@neighbours0);
    $self->add_edge (@neighbours1);
    $self->set_edge_attribute (@neighbours0, 'heat', 1);
    $self->set_edge_attribute (@neighbours1, 'heat', 1);
    $self->delete_vertices (@todie);

    return 1;
}

# Find a low crinkliness polygon and split it into two.

sub Sow
{
    my $self = shift;
    my $crinkliness = shift || 0.0133;
    my $loops = $self->Loops;
    for my $loop (@{$loops})
    {
        my $polygon = $self->Polygon ($loop);
        next unless $polygon->crinkliness < $crinkliness;
        next if $polygon->isClockwise;
        my $offset = int (scalar @{$loop} / 2);
        print STDERR "Subdividing polygon with crinkliness ". $polygon->crinkliness . "(< $crinkliness)\n";
        my $distances;
        for my $id_vertex (0 .. scalar @{$loop} - $offset - 2)
        {
            $distances->{$id_vertex} = $self->Get_Distance ($id_vertex, $id_vertex + $offset +1)
                                     + $self->Get_Distance ($id_vertex +1, $id_vertex + $offset -1);
        }
        my @sorted = sort {$distances->{$a} <=> $distances->{$b}} keys %{$distances};
        my $victim = shift @sorted;
        $self->Split_Edges ($loop->[$victim], $loop->[$victim +1], $loop->[$victim + $offset], $loop->[$victim + $offset +1]);
        last;
    }
}

# split two edges, adding two new vertices, and add a new edge between these vertices

sub Split_Edges
{
    my $self = shift;
    my @todie = @_;
    my $new0 = $self->Next;
    $self->add_vertex ($new0);
    my $new1 = $self->Next;
    $self->add_vertex ($new1);

    $self->add_edge ($new0, $new1);
    $self->set_edge_attribute ($new0, $new1, 'heat', 1);

    $self->add_edge ($new0, $todie[0]);
    $self->set_edge_attribute ($new0, $todie[0], 'heat', 1);

    $self->add_edge ($new0, $todie[1]);
    $self->set_edge_attribute ($new0, $todie[1], 'heat', 1);

    $self->add_edge ($new1, $todie[2]);
    $self->set_edge_attribute ($new1, $todie[2], 'heat', 1);

    $self->add_edge ($new1, $todie[3]);
    $self->set_edge_attribute ($new1, $todie[3], 'heat', 1);

    $self->delete_edge ($todie[0], $todie[1]);
    $self->delete_edge ($todie[2], $todie[3]);

    my $coor0 = _add_vectors ($self->get_vertex_attribute ($todie[0], 'coor'), $self->get_vertex_attribute ($todie[1], 'coor'));
    $self->set_vertex_attribute ($new0, 'coor', _scale_vector ($coor0, 0.5));

    my $coor1 = _add_vectors ($self->get_vertex_attribute ($todie[2], 'coor'), $self->get_vertex_attribute ($todie[3], 'coor'));
    $self->set_vertex_attribute ($new1, 'coor', _scale_vector ($coor1, 0.5));
}

# find an unused id for a new vertex

sub Next
{
    my $self = shift;
    my $index = 0;
    my $waiting = 1;
    while ($waiting)
    {
        $index++;
        $waiting = 0 unless grep /^$index$/, $self->vertices;
    }
    return $index;
}

# retrieve left-hand angle between three named points regardless of edges

sub Angle
{
    my $self = shift;
    my ($id_a, $id_b, $id_c) = @_;

    my $coor_a = $self->get_vertex_attribute ($id_a, 'coor');
    my $coor_b = $self->get_vertex_attribute ($id_b, 'coor');
    my $coor_c = $self->get_vertex_attribute ($id_c, 'coor');

    my $vec_ab = [$coor_b->[0] - $coor_a->[0],
                  $coor_b->[1] - $coor_a->[1],
                  $coor_b->[2] - $coor_a->[2]];
    my $vec_bc = [$coor_c->[0] - $coor_b->[0],
                  $coor_c->[1] - $coor_b->[1],
                  $coor_c->[2] - $coor_b->[2]];

    $vec_ab->[0] = 0.00000000001 if $vec_ab->[0] == 0;
    $vec_bc->[0] = 0.00000000001 if $vec_bc->[0] == 0;

    my $bearing_ab = atan ($vec_ab->[1] / $vec_ab->[0]);
    $bearing_ab += $PI if $vec_ab->[0] < 0;

    my $bearing_bc = atan ($vec_bc->[1] / $vec_bc->[0]);
    $bearing_bc += $PI if $vec_bc->[0] < 0;

    my $angle = $PI - $bearing_bc + $bearing_ab;
    $angle -= $PI2 if $angle >= $PI2;
    $angle += $PI2 if $angle < 0.0;

    return $angle;
}

# retrieve all closed loops with no internal vertices from the graph
# these will be anticlockwise except for the outer perimeter which will be clockwise

sub Loops
{
    my $self = shift;
    my $seen_links = {};
    my $loops = [];
    my @edges = map {$_, [$_->[1], $_->[0]]} $self->edges;
    for my $edge (@edges)
    {
        my $looped = 0;
        my $loop = $edge;
        my $identifier = join ':', @{$edge};
        next if defined $seen_links->{$identifier};
        $seen_links->{$identifier} = 1;
        my $going = 1;
        while ($going)
        {
            my $next = $self->Next_Left ($loop->[-2], $loop->[-1]);
            $going = 0 unless defined $next;
            next unless $going;
            $going = 0 if defined $seen_links->{join ':', $loop->[-1], $next};
            $going = 0 if $loop->[-1] eq $next;
            next unless $going;
            $seen_links->{join ':', $loop->[-1], $next} = 1;
            push @{$loop}, $next;
        }
        push @{$loops}, $loop;
    }
    my @loops_real;
    for my $loop (@{$loops})
    {
        next unless $loop->[-1] eq $loop->[0];
        pop @{$loop};
        push @loops_real, $loop;
    }
    return [@loops_real];
}

# Return a Math::Polygon object corresponding to the named vertices.
# If last vertex is the same as the first then they will be merged.

sub Polygon
{
    my $self = shift;
    my $vertices = shift;
    push @{$vertices}, $vertices->[0] unless ($vertices->[0] eq $vertices->[-1]);
    return Math::Polygon->new (map {$self->get_vertex_attribute ($_, 'coor')} @{$vertices});
}

# calculate crinkliness, large blocks have low crinkliness, thin blocks have high crinkliness

sub Math::Polygon::crinkliness
{
    my $polygon = shift;
    $polygon->perimeter / $polygon->area;
}

# Continue path from a to b to c, turning left when possible

sub Next_Left
{
    my $self = shift;
    my ($m, $n) = @_;
    my @neighbours = $self->neighbours ($n);
    @neighbours = grep !/^$m$/, @neighbours;
    return undef unless scalar @neighbours;
    @neighbours = sort {$self->Angle ($m, $n, $a) <=> $self->Angle ($m, $n, $b)} @neighbours;
    return $neighbours[0];
}


sub _distance
{
    my ($a, $b) = @_;
    sqrt (($a->[0] - $b->[0])**2 + ($a->[1] - $b->[1])**2 + ($a->[2] - $b->[2])**2);
}

sub _scale_vector
{
    my $vector = shift;
    my $scale = shift;
    return [$vector->[0] * $scale, $vector->[1] * $scale, $vector->[2] * $scale];
}

sub _add_vectors
{
    my @vectors = @_;
    my $result = [0,0,0];
    for (@vectors)
    {
        $result->[0] += $_->[0];
        $result->[1] += $_->[1];
        $result->[2] += $_->[2];
    }
    return $result;
}

1;
